from collections.abc import Iterable
from typing import Any, Literal, cast

from haiway import ctx

from draive.generation.model.types import ModelGeneratorDecoder
from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMToolRequests,
    LMMToolResponses,
)
from draive.multimodal import (
    Multimodal,
    MultimodalContent,
)
from draive.parameters import DataModel
from draive.prompts import Prompt
from draive.tools import Tool, Toolbox

__all__ = ("generate_model",)


async def generate_model[Generated: DataModel](  # noqa: C901, PLR0912
    generated: type[Generated],
    /,
    *,
    instruction: Instruction | str,
    input: Prompt | Multimodal,  # noqa: A002
    schema_injection: Literal["auto", "full", "simplified", "skip"],
    tools: Toolbox | Iterable[Tool] | None,
    examples: Iterable[tuple[Multimodal, Generated]] | None,
    decoder: ModelGeneratorDecoder | None,
    **extra: Any,
) -> Generated:
    with ctx.scope("generate_model"):
        toolbox: Toolbox = Toolbox.of(tools)

        generation_instruction: Instruction
        match instruction:
            case str(instruction):
                generation_instruction = Instruction.of(instruction)

            case Instruction() as instruction:
                generation_instruction = instruction

        extended_instruction: Instruction
        match schema_injection:
            case "auto":
                extended_instruction = generation_instruction.extended(
                    DEFAULT_INSTRUCTION_EXTENSION.format(
                        schema=generated.simplified_schema(indent=2),
                    ),
                    joiner="\n\n",
                )

            case "full":
                extended_instruction = generation_instruction.updated(
                    schema=generated.json_schema(indent=2),
                )

            case "simplified":
                extended_instruction = generation_instruction.updated(
                    schema=generated.simplified_schema(indent=2),
                )

            case "skip":
                extended_instruction = Instruction.of(
                    generation_instruction,
                )

        context: list[LMMContextElement] = [
            *[
                message
                for example in examples or []
                for message in [
                    LMMInput.of(example[0]),
                    LMMCompletion.of(example[1].to_json(indent=2)),
                ]
            ],
        ]

        match input:
            case Prompt() as prompt:
                context.extend(prompt.content)

            case value:
                context.append(LMMInput.of(value))

        recursion_level: int = 0
        while recursion_level <= toolbox.repeated_calls_limit:
            match await LMM.completion(
                instruction=extended_instruction,
                context=context,
                tool_selection=toolbox.tool_selection(repetition_level=recursion_level),
                tools=toolbox.available_tools(),
                output=generated,  # provide model specification
                **extra,
            ):
                case LMMCompletion() as completion:
                    if decoder := decoder:
                        return generated.from_mapping(decoder(completion.content))

                    elif (artifacts := completion.content.artifacts()) and isinstance(
                        artifacts[0], generated
                    ):
                        return cast(Generated, artifacts[0])

                    else:
                        return generated.from_json(completion.content.to_str())

                case LMMToolRequests() as tool_requests:
                    context.append(tool_requests)
                    tool_responses: LMMToolResponses = await toolbox.respond_all(tool_requests)

                    if direct_responses := [
                        response
                        for response in tool_responses.responses
                        if response.handling == "direct_result"
                    ]:
                        for response in direct_responses:
                            if isinstance(response, generated):
                                # return first response matching requested model
                                return response

                            else:
                                continue

                        direct_responses_content: MultimodalContent = MultimodalContent.of(
                            *[response.content for response in direct_responses]
                        )

                        # TODO: check if this join makes any sense,
                        # perhaps we could merge json objects instead?
                        if decoder := decoder:
                            return generated.from_mapping(decoder(direct_responses_content))

                        else:
                            return generated.from_json(direct_responses_content.to_str())

                    else:
                        context.append(tool_responses)

            recursion_level += 1  # continue with next recursion level

        raise RuntimeError("LMM exceeded limit of recursive calls")


DEFAULT_INSTRUCTION_EXTENSION: str = """\
<FORMAT>
Provide the result using a single raw valid JSON object that adheres strictly to the given \
SCHEMA without any comments, formatting, or additional elements.
<SCHEMA>
{schema}
</SCHEMA>
</FORMAT>
"""



from collections.abc import Iterable
from typing import Any, Literal, Protocol, runtime_checkable

from draive.instructions import Instruction
from draive.multimodal import Multimodal, MultimodalContent
from draive.parameters import DataModel
from draive.prompts import Prompt
from draive.tools import Tool, Toolbox

__all__ = (
    "ModelGenerating",
    "ModelGeneratorDecoder",
)


@runtime_checkable
class ModelGeneratorDecoder(Protocol):
    def __call__(
        self,
        generated: MultimodalContent,
    ) -> dict[str, Any]: ...


@runtime_checkable
class ModelGenerating(Protocol):
    async def __call__[Generated: DataModel](
        self,
        generated: type[Generated],
        /,
        *,
        instruction: Instruction | str,
        input: Prompt | Multimodal,  # noqa: A002
        schema_injection: Literal["auto", "full", "simplified", "skip"],
        tools: Toolbox | Iterable[Tool] | None,
        examples: Iterable[tuple[Multimodal, Generated]] | None,
        decoder: ModelGeneratorDecoder | None,
        **extra: Any,
    ) -> Generated: ...



from draive.guardrails.moderation.state import GuardrailsModeration
from draive.guardrails.moderation.types import (
    GuardrailsInputModerationException,
    GuardrailsModerationChecking,
    GuardrailsModerationException,
    GuardrailsOutputModerationException,
)

__all__ = (
    "GuardrailsInputModerationException",
    "GuardrailsModeration",
    "GuardrailsModerationChecking",
    "GuardrailsModerationException",
    "GuardrailsOutputModerationException",
)



from asyncio import gather
from collections.abc import Callable, Sequence
from itertools import chain
from typing import Any, cast

from haiway import ObservabilityLevel, State, as_list, ctx
from openai.types.create_embedding_response import CreateEmbeddingResponse

from draive.embedding import Embedded, TextEmbedding
from draive.openai.api import OpenAIAPI
from draive.openai.config import OpenAIEmbeddingConfig
from draive.openai.utils import unwrap_missing
from draive.parameters import DataModel

__all__ = ("OpenAIEmbedding",)


class OpenAIEmbedding(OpenAIAPI):
    def text_embedding(self) -> TextEmbedding:
        """
        Prepare TextEmbedding implementation using OpenAI service.
        """
        return TextEmbedding(embedding=self.create_texts_embedding)

    async def create_texts_embedding[Value: DataModel | State](
        self,
        values: Sequence[Value] | Sequence[str],
        /,
        attribute: Callable[[Value], str] | None = None,
        *,
        config: OpenAIEmbeddingConfig | None = None,
        **extra: Any,
    ) -> Sequence[Embedded[Value]] | Sequence[Embedded[str]]:
        """
        Create texts embedding with OpenAI embedding service.
        """
        embedding_config: OpenAIEmbeddingConfig = config or ctx.state(OpenAIEmbeddingConfig)
        with ctx.scope("openai_text_embedding", embedding_config):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "embedding.provider": "openai",
                    "embedding.model": embedding_config.model,
                    "embedding.dimensions": embedding_config.dimensions,
                },
            )
            attributes: list[str]
            if attribute is None:
                attributes = cast(list[str], as_list(values))

            else:
                attributes = [attribute(cast(Value, value)) for value in values]

            assert all(isinstance(element, str) for element in attributes)  # nosec: B101

            responses: list[CreateEmbeddingResponse] = await gather(
                *[
                    self._client.embeddings.create(
                        input=attributes[index : index + embedding_config.batch_size],
                        model=embedding_config.model,
                        dimensions=unwrap_missing(embedding_config.dimensions),
                        encoding_format="float",
                        timeout=unwrap_missing(embedding_config.timeout),
                    )
                    for index in range(0, len(attributes), embedding_config.batch_size)
                ]
            )

            return cast(
                Sequence[Embedded[Value]] | Sequence[Embedded[str]],
                [
                    Embedded(
                        value=value,
                        vector=embedding.embedding,
                    )
                    for value, embedding in zip(
                        values,
                        chain.from_iterable([response.data for response in responses]),
                        strict=True,
                    )
                ],
            )



from draive.resources.state import Resources
from draive.resources.template import ResourceTemplate, resource
from draive.resources.types import (
    Resource,
    ResourceContent,
    ResourceDeclaration,
    ResourceException,
    ResourceFetching,
    ResourceListFetching,
    ResourceMissing,
    ResourceTemplateDeclaration,
    ResourceUploading,
)

__all__ = (
    "Resource",
    "ResourceContent",
    "ResourceDeclaration",
    "ResourceException",
    "ResourceFetching",
    "ResourceListFetching",
    "ResourceMissing",
    "ResourceTemplate",
    "ResourceTemplateDeclaration",
    "ResourceUploading",
    "Resources",
    "resource",
)



from collections.abc import Sequence

from haiway import MISSING, Missing

from draive.configuration import Config

__all__ = (
    "MistralChatConfig",
    "MistralEmbeddingConfig",
)


class MistralChatConfig(Config):
    model: str
    temperature: float = 1.0
    top_p: float | Missing = MISSING
    seed: int | Missing = MISSING
    max_tokens: int | Missing = MISSING
    timeout: float | Missing = MISSING
    stop_sequences: Sequence[str] | Missing = MISSING


class MistralEmbeddingConfig(Config):
    model: str
    batch_size: int = 128



from draive import MediaReference, MultimodalContent, MultimodalTagElement


def test_returns_none_with_empty():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.empty,
        )
        is None
    )


def test_returns_none_without_tag():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("Lorem ipsum"),
        )
        is None
    )

    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of(
                "Lorem",
                MediaReference.of("http://image", media="image/png"),
                "ipsum",
            ),
        )
        is None
    )


def test_returns_none_with_other_tag():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("<other>Lorem ipsum</other>"),
        )
        is None
    )

    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of(
                "<other>Lorem",
                MediaReference.of("http://image", media="image/png"),
                "ipsum</other>",
            ),
        )
        is None
    )


def test_returns_none_with_closing_tag():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("Lorem ipsum</test>"),
        )
        is None
    )

    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of(
                "Lorem",
                MediaReference.of("http://image", media="image/png"),
                "ipsum</test>",
            ),
        )
        is None
    )


def test_returns_empty_with_closed_tag():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of("Lorem ipsum<test/>"),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.empty,
    )

    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of(
            "Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum<test/>",
        ),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.empty,
    )


def test_returns_none_with_reversed_tags():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("</test>Lorem ipsum<test>"),
        )
        is None
    )

    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of(
                "</test>Lorem",
                MediaReference.of("http://image", media="image/png"),
                "ipsum<test>",
            ),
        )
        is None
    )


def test_returns_none_without_closing_tag():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("<test>Lorem ipsum"),
        )
        is None
    )

    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of(
                "<test>Lorem",
                MediaReference.of("http://image", media="image/png"),
                "ipsum",
            ),
        )
        is None
    )


def test_returns_none_with_malformed_opening_tag():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("<testx>Lorem ipsum</test>"),
        )
        is None
    )

    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of(
                "<testx>Lorem",
                MediaReference.of("http://image", media="image/png"),
                "ipsum</test>",
            ),
        )
        is None
    )


def test_returns_none_with_malformed_closing_tag():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("<test>Lorem ipsum</testx>"),
        )
        is None
    )

    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of(
                "<test>Lorem",
                MediaReference.of("http://image", media="image/png"),
                "ipsum</testx>",
            ),
        )
        is None
    )


def test_returns_content_with_valid_tag():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</test>"),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("Lorem ipsum"),
    )

    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of(
            "<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
        ),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of(
            "Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
    )


def test_returns_content_with_surrounded_tag():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of("Lorem<test>Lorem ipsum</test>ipsum"),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("Lorem ipsum"),
    )

    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of(
            MediaReference.of("http://image", media="image/png"),
            "Lorem<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>ipsum",
            MediaReference.of("http://image", media="image/png"),
        ),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of(
            "Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
    )


def test_returns_content_with_opening_tag_containing_attributes():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of('Lorem<test key="value">Lorem ipsum</test>ipsum'),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("Lorem ipsum"),
        attributes={"key": "value"},
    )

    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of('Lorem<test key="value" next="more">Lorem ipsum</test>ipsum'),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("Lorem ipsum"),
        attributes={"key": "value", "next": "more"},
    )

    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of(
            'Lorem<test key="value" next="more" spaces="s p a ce" >Lorem ipsum</test>ipsum'
        ),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("Lorem ipsum"),
        attributes={"key": "value", "next": "more", "spaces": "s p a ce"},
    )

    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of(
            'Lorem<test key="value" escaping="\\"escape" spaces="s p a ce" >Lorem ipsum</test>ipsum'
        ),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("Lorem ipsum"),
        attributes={"key": "value", "escaping": '"escape', "spaces": "s p a ce"},
    )
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of(
            MediaReference.of("http://image", media="image/png"),
            'Lorem<test key="value">Lorem',
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>ipsum",
            MediaReference.of("http://image", media="image/png"),
        ),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of(
            "Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
        attributes={"key": "value"},
    )


def test_returns_none_with_opening_tag_containing_malformed_attributes():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of('Lorem<test key=value">Lorem ipsum</test>ipsum'),
        )
        is None
    )
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of('Lorem<testkey="value">Lorem ipsum</test>ipsum'),
        )
        is None
    )
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of('Lorem<test key="va\nlue">Lorem ipsum</test>ipsum'),
        )
        is None
    )
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("Lorem<testkey=>Lorem ipsum</test>ipsum"),
        )
        is None
    )


def test_returns_first_content_with_multiple_tags():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</test><test>Other</test>"),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("Lorem ipsum"),
    )

    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of(
            "<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test><test>Other</test>",
        ),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of(
            "Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
    )


def test_returns_outer_content_with_multiple_nested_tags():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of("<test>Other<test>Lorem ipsum</test></test>"),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("Other<test>Lorem ipsum"),
    )

    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of(
            "<test>Other<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test></test>",
        ),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of(
            "Other<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
    )


def test_returns_nested_content_with_fake_tags():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of("<test>Lorem<ipsum</test>"),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("Lorem<ipsum"),
    )

    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of(
            "<test>Lorem<",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
        ),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of(
            "Lorem<",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
    )


def test_returns_none_with_nested_in_other_tags():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of(
                "<other>Other<more><test>Lorem</more>ipsum</test></other>"
            ),
        )
        is None
    )

    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of(
                "<other>Other<more><test>Lorem</more>",
                MediaReference.of("http://image", media="image/png"),
                "ipsum</test></other>",
            ),
        )
        is None
    )


def test_returns_content_with_multiple_tags():
    assert list(
        MultimodalTagElement.parse(
            content=MultimodalContent.of(
                "<test>Lorem ipsum</test><other>Dolor</other><another>Sit amet</another>"
            ),
        )
    ) == [
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of("Lorem ipsum"),
        ),
        MultimodalTagElement(
            name="other",
            content=MultimodalContent.of("Dolor"),
        ),
        MultimodalTagElement(
            name="another",
            content=MultimodalContent.of("Sit amet"),
        ),
    ]

    assert list(
        MultimodalTagElement.parse(
            content=MultimodalContent.of(
                "<test>Lorem ",
                MediaReference.of("http://image", media="image/png"),
                "ipsum</test><other>Dolor</other><another>Sit ",
                MediaReference.of("http://image", media="image/png"),
                "amet</another>",
            ),
        )
    ) == [
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of(
                "Lorem ",
                MediaReference.of("http://image", media="image/png"),
                "ipsum",
            ),
        ),
        MultimodalTagElement(
            name="other",
            content=MultimodalContent.of(
                "Dolor",
            ),
        ),
        MultimodalTagElement(
            name="another",
            content=MultimodalContent.of(
                "Sit ",
                MediaReference.of("http://image", media="image/png"),
                "amet",
            ),
        ),
    ]


def test_returns_content_with_multiple_filtered_tags():
    assert list(
        MultimodalTagElement.parse(
            "test",
            content=MultimodalContent.of(
                "<test>Lorem ipsum</test><test>Dolor</test><test>Sit amet</test>"
            ),
        )
    ) == [
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of("Lorem ipsum"),
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of("Dolor"),
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of("Sit amet"),
        ),
    ]

    assert list(
        MultimodalTagElement.parse(
            "test",
            content=MultimodalContent.of(
                "<test>Lorem ",
                MediaReference.of("http://image", media="image/png"),
                "ipsum</test><test>Dolor</test><test>Sit ",
                MediaReference.of("http://image", media="image/png"),
                "amet</test>",
            ),
        )
    ) == [
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of(
                "Lorem ",
                MediaReference.of("http://image", media="image/png"),
                "ipsum",
            ),
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of(
                "Dolor",
            ),
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of(
                "Sit ",
                MediaReference.of("http://image", media="image/png"),
                "amet",
            ),
        ),
    ]


def test_skips_content_from_different_filtered_tags():
    assert list(
        MultimodalTagElement.parse(
            "test",
            content=MultimodalContent.of(
                "<test>Lorem ipsum</test><other>Other</other><test>Dolor</test>"
            ),
        )
    ) == [
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of("Lorem ipsum"),
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of("Dolor"),
        ),
    ]

    assert list(
        MultimodalTagElement.parse(
            "test",
            content=MultimodalContent.of(
                "<test>Lorem ",
                MediaReference.of("http://image", media="image/png"),
                "ipsum</test><other>Other</other><test>Sit ",
                MediaReference.of("http://image", media="image/png"),
                "amet</test>",
            ),
        )
    ) == [
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of(
                "Lorem ",
                MediaReference.of("http://image", media="image/png"),
                "ipsum",
            ),
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of(
                "Sit ",
                MediaReference.of("http://image", media="image/png"),
                "amet",
            ),
        ),
    ]


### generated ###


def test_handles_empty_tag_content():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of("<test></test>"),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.empty,
    )


def test_handles_whitespace_only_tag_content():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of("<test>   \n\t  </test>"),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("   \n\t  "),
    )


def test_handles_special_characters_in_attributes():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of('<test special="&quot;\'<>/">\ntest\n</test>'),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("\ntest\n"),
        attributes={"special": "&quot;'<>/"},
    )


def test_handles_escaped_quotes_in_attributes():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of('<test quote="\\"quoted\\"">content</test>'),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("content"),
        attributes={"quote": '"quoted"'},
    )


def test_handles_multiple_consecutive_tags():
    assert list(
        MultimodalTagElement.parse(
            "test",
            content=MultimodalContent.of("<test/><test/><test/>"),
        )
    ) == [
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.empty,
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.empty,
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.empty,
        ),
    ]


def test_handles_mixed_self_closing_and_normal_tags():
    assert list(
        MultimodalTagElement.parse(
            "test",
            content=MultimodalContent.of("<test/>text<test>content</test><test/>"),
        )
    ) == [
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.empty,
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.of("content"),
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.empty,
        ),
    ]


def test_handles_tag_names_with_numbers():
    assert MultimodalTagElement.parse_first(
        "test123",
        content=MultimodalContent.of("<test123>content</test123>"),
    ) == MultimodalTagElement(
        name="test123",
        content=MultimodalContent.of("content"),
    )


def test_handles_invalid_attribute_formats():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of('<test a="1" b= c d="2">content</test>'),
        )
        is None
    )


def test_handles_unclosed_attribute_quotes():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of('<test attr="value>content</test>'),
        )
        is None
    )


def test_handles_nested_angle_brackets():
    assert MultimodalTagElement.parse_first(
        "test",
        content=MultimodalContent.of("<test>a < b > c</test>"),
    ) == MultimodalTagElement(
        name="test",
        content=MultimodalContent.of("a < b > c"),
    )


def test_handles_multiple_spaces_in_tags():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("<test   key='value'   >content</test>"),
        )
        is None  # This case should be handled in future
    )


def test_handles_malformed_closing_tags():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("<test>content</test attr='value'>"),
        )
        is None
    )


def test_handles_case_sensitivity():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("<TEST>content</TEST>"),
        )
        is None
    )


def test_handles_mixed_media_and_empty_tags():
    assert list(
        MultimodalTagElement.parse(
            content=MultimodalContent.of(
                "<test/>",
                MediaReference.of("http://image1", media="image/png"),
                "<test></test>",
                MediaReference.of("http://image2", media="image/png"),
                "<test/>",
            ),
        )
    ) == [
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.empty,
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.empty,
        ),
        MultimodalTagElement(
            name="test",
            content=MultimodalContent.empty,
        ),
    ]


def test_handles_attribute_without_value():
    assert (
        MultimodalTagElement.parse_first(
            "test",
            content=MultimodalContent.of("<test solo>content</test>"),
        )
        is None
    )



from collections.abc import AsyncIterator, Iterable
from typing import Any, Literal, overload

from haiway import ObservabilityLevel, asynchronous, ctx

from draive.bedrock.api import BedrockAPI
from draive.bedrock.config import BedrockChatConfig
from draive.bedrock.lmm import (
    convert_context_element,
    output_as_response_declaration,
    tools_as_tool_config,
)
from draive.bedrock.models import ChatCompletionResponse, ChatMessage
from draive.bedrock.types import BedrockException
from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContext,
    LMMOutput,
    LMMOutputSelection,
    LMMStreamOutput,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.multimodal import MediaData, MultimodalContent, MultimodalContentElement, TextContent


class BedrockLMMGeneration(BedrockAPI):
    def lmm(self) -> LMM:
        return LMM(completing=self.lmm_completion)

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        config: BedrockChatConfig | None = None,
        output: LMMOutputSelection,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        config: BedrockChatConfig | None = None,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    async def lmm_completion(  # noqa: C901, PLR0912
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        config: BedrockChatConfig | None = None,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput:
        completion_config: BedrockChatConfig = config or ctx.state(BedrockChatConfig)
        with ctx.scope("bedrock_lmm_completion", completion_config):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "lmm.provider": "bedrock",
                    "lmm.model": completion_config.model,
                    "lmm.temperature": completion_config.temperature,
                    "lmm.max_tokens": completion_config.max_tokens,
                    "lmm.tools": [tool["name"] for tool in tools] if tools else None,
                    "lmm.tool_selection": f"{tool_selection}" if tools else None,
                    "lmm.stream": stream,
                    "lmm.output": f"{output}",
                    "lmm.instruction": f"{instruction}",
                    "lmm.context": [element.to_str() for element in context],
                },
            )

            if stream:
                raise NotImplementedError("bedrock streaming is not implemented yet")

            output_decoder = output_as_response_declaration(output)

            messages: list[ChatMessage] = [convert_context_element(element) for element in context]

            tool_config = tools_as_tool_config(tools, tool_selection=tool_selection)

            completion: ChatCompletionResponse = await self._chat_completion(
                config=completion_config,
                instruction=Instruction.formatted(instruction),
                messages=messages,
                tool_config=tool_config,
            )

            ctx.record(
                ObservabilityLevel.INFO,
                metric="lmm.input_tokens",
                value=completion["usage"]["inputTokens"],
                unit="tokens",
                attributes={"lmm.model": completion_config.model},
            )
            ctx.record(
                ObservabilityLevel.INFO,
                metric="lmm.output_tokens",
                value=completion["usage"]["outputTokens"],
                unit="tokens",
                attributes={"lmm.model": completion_config.model},
            )

            message_parts: list[MultimodalContentElement] = []
            tool_calls: list[LMMToolRequest] = []
            for part in completion["output"]["message"]["content"]:
                match part:
                    case {"text": str() as text}:
                        message_parts.append(TextContent(text=text))

                    case {
                        "image": {
                            "format": str() as data_format,
                            "source": {"bytes": bytes() as data},
                        }
                    }:
                        media_type: Any
                        match data_format:
                            case "png":
                                media_type = "image/png"

                            case "jpeg":
                                media_type = "image/jpeg"

                            case "gif":
                                media_type = "image/gif"

                            case _:  # pyright: ignore[reportUnnecessaryComparison]
                                media_type = "image"

                        message_parts.append(
                            MediaData.of(
                                data,
                                media=media_type,
                            )
                        )

                    case {
                        "toolUse": {
                            "toolUseId": str() as identifier,
                            "name": str() as tool,
                            "input": arguments,
                        }
                    }:
                        tool_calls.append(
                            LMMToolRequest(
                                identifier=identifier,
                                tool=tool,
                                arguments=arguments,
                            )
                        )

                    case _:
                        pass

            match completion["stopReason"]:
                case "end_turn" | "stop_sequence":
                    message_completion = LMMCompletion.of(
                        output_decoder(MultimodalContent.of(*message_parts))
                    )

                    ctx.record(
                        ObservabilityLevel.INFO,
                        event="lmm.completion",
                    )
                    return message_completion

                case "tool_use":
                    tools_completion = LMMToolRequests(requests=tool_calls)
                    ctx.record(
                        ObservabilityLevel.INFO,
                        event="lmm.tool_requests",
                        attributes={"lmm.tools": [call.tool for call in tool_calls]},
                    )
                    return tools_completion

                case _:
                    raise BedrockException("Invalid Bedrock response")

    @asynchronous
    def _chat_completion(
        self,
        *,
        config: BedrockChatConfig,
        instruction: str | None,
        messages: list[ChatMessage],
        tool_config: dict[str, Any] | None,
    ) -> ChatCompletionResponse:
        parameters: dict[str, Any] = {
            "modelId": config.model,
            "messages": messages,
            "inferenceConfig": {
                "temperature": config.temperature,
            },
        }

        if instruction:
            parameters["system"] = [
                {
                    "text": instruction,
                }
            ]

        if tool_config:
            parameters["toolConfig"] = tool_config
        if config.max_tokens:
            parameters["inferenceConfig"]["maxTokens"] = config.max_tokens
        if config.top_p:
            parameters["inferenceConfig"]["topP"] = config.top_p
        if config.stop_sequences:
            parameters["inferenceConfig"]["stopSequences"] = config.stop_sequences

        return self._client.converse(**parameters)



from collections.abc import Callable, Iterable
from typing import Any, Literal, cast

from draive.bedrock.models import ChatMessage, ChatMessageContent, ChatTool
from draive.lmm import (
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMToolRequests,
    LMMToolResponses,
    LMMToolSpecification,
)
from draive.lmm.types import LMMOutputSelection, LMMToolSelection
from draive.multimodal import (
    MediaData,
    MediaReference,
    MultimodalContentElement,
    TextContent,
)
from draive.multimodal.content import MultimodalContent
from draive.parameters import DataModel

__all__ = (
    "convert_content_element",
    "convert_context_element",
    "convert_tool",
)


def convert_content_element(
    element: MultimodalContentElement,
) -> ChatMessageContent:
    match element:
        case TextContent() as text:
            return {"text": text.text}

        case MediaData() as media_data:
            if media_data.kind != "image":
                raise ValueError("Unsupported message content", media_data)

            image_format: Literal["png", "jpeg", "gif"]
            match media_data.media:
                case "image/png":
                    image_format = "png"

                case "image/jpeg":
                    image_format = "jpeg"

                case "image/gif":
                    image_format = "gif"

                case _:
                    raise ValueError("Unsupported message content", media_data)

            return {
                "image": {
                    "format": image_format,
                    "source": {"bytes": media_data.data},
                }
            }

        case MediaReference() as media_reference:
            raise ValueError("Unsupported message content", media_reference)

        case DataModel() as data:
            return {"text": data.to_json()}


def convert_context_element(
    element: LMMContextElement,
) -> ChatMessage:
    match element:
        case LMMInput() as input:
            return ChatMessage(
                role="user",
                content=[convert_content_element(part) for part in input.content.parts],
            )

        case LMMCompletion() as completion:
            return ChatMessage(
                role="assistant",
                content=[convert_content_element(part) for part in completion.content.parts],
            )

        case LMMToolRequests() as requests:
            return ChatMessage(
                role="assistant",
                content=[
                    {
                        "toolUse": {
                            "toolUseId": request.identifier,
                            "name": request.tool,
                            "input": request.arguments,
                        }
                    }
                    for request in requests.requests
                ],
            )

        case LMMToolResponses() as tool_responses:
            return ChatMessage(
                role="user",
                content=[
                    {
                        "toolResult": {
                            "toolUseId": response.identifier,
                            "content": [
                                cast(Any, convert_content_element(part))
                                for part in response.content.parts
                            ],
                            "status": "error" if response.handling == "error" else "success",
                        },
                    }
                    for response in tool_responses.responses
                ],
            )


def convert_tool(tool: LMMToolSpecification) -> ChatTool:
    return {
        "name": tool["name"],
        "description": tool["description"] or "",
        "inputSchema": {"json": tool["parameters"]},
    }


def output_as_response_declaration(
    output: LMMOutputSelection,
) -> Callable[[MultimodalContent], MultimodalContent]:
    match output:
        case "auto":
            return _auto_output_conversion

        case ["text"] | "text":
            return _auto_output_conversion

        case "json":
            return _json_output_conversion

        case "image":
            raise NotImplementedError("image output is not supported yet")

        case "audio":
            raise NotImplementedError("audio output is not supported yet")

        case "video":
            raise NotImplementedError("video output is not supported yet")

        case [*_]:
            raise NotImplementedError("multimodal output is not supported yet")

        case model:
            return _prepare_model_output_conversion(model)


def _auto_output_conversion(
    output: MultimodalContent,
    /,
) -> MultimodalContent:
    return output


def _text_output_conversion(
    output: MultimodalContent,
    /,
) -> MultimodalContent:
    return MultimodalContent.of(output.to_str())


def _audio_output_conversion(
    output: MultimodalContent,
    /,
) -> MultimodalContent:
    return MultimodalContent.of(*output.media("audio"))


def _json_output_conversion(
    output: MultimodalContent,
    /,
) -> MultimodalContent:
    return MultimodalContent.of(DataModel.from_json(output.to_str()))


def _prepare_model_output_conversion(
    model: type[DataModel],
    /,
) -> Callable[[MultimodalContent], MultimodalContent]:
    def _model_output_conversion(
        output: MultimodalContent,
        /,
    ) -> MultimodalContent:
        return MultimodalContent.of(model.from_json(output.to_str()))

    return _model_output_conversion


def tools_as_tool_config(
    tools: Iterable[LMMToolSpecification] | None,
    /,
    *,
    tool_selection: LMMToolSelection,
) -> dict[str, Any] | None:
    toolChoice: dict[str, Any]
    match tool_selection:
        case "auto":
            toolChoice = {"auto": {}}

        case "required":
            toolChoice = {"any": {}}

        case "none":
            return None

        case tool:
            toolChoice = {"tool": {"name": tool}}

    tools_list = [{"toolSpec": convert_tool(tool)} for tool in tools or ()]
    if not tools_list:
        return None

    return {
        "tools": tools_list,
        "toolChoice": toolChoice,
    }



# LMMSession
from asyncio import Task, TaskGroup
from collections.abc import AsyncIterator, Sequence
from contextlib import AbstractAsyncContextManager
from itertools import chain
from typing import Any

from google.genai.live import AsyncSession
from google.genai.types import (
    Content,
    ContentDict,
    FunctionCallingConfigMode,
    FunctionDeclarationDict,
    Modality,
    SchemaDict,
    SpeechConfigDict,
)
from haiway import AsyncQueue, as_list, ctx

from draive.gemini.api import GeminiAPI
from draive.gemini.config import GeminiLiveConfig
from draive.gemini.lmm import (
    content_element_as_part,
    context_element_as_content,
    output_as_response_declaration,
    result_part_as_content_or_call,
    tools_as_tools_config,
)
from draive.gemini.utils import unwrap_missing
from draive.instructions import Instruction
from draive.lmm import (
    LMMContext,
    LMMSession,
    LMMSessionOutput,
    LMMSessionOutputSelection,
    LMMStreamInput,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.lmm.types import LMMSessionEvent, LMMStreamChunk, LMMToolRequest, LMMToolResponse
from draive.multimodal import MediaData, MediaReference
from draive.multimodal.content import MultimodalContent
from draive.multimodal.text import TextContent
from draive.parameters import DataModel

__all__ = ("GeminiLMMSession",)


class GeminiLMMSession(GeminiAPI):
    def lmm_session(self) -> LMMSession:
        return LMMSession(preparing=self.lmm_session_prepare)

    async def lmm_session_prepare(
        self,
        *,
        instruction: Instruction | None = None,
        initial_context: LMMContext | None = None,
        input_stream: AsyncIterator[LMMStreamInput],
        output: LMMSessionOutputSelection = "auto",
        tools: Sequence[LMMToolSpecification] | None = None,
        tool_selection: LMMToolSelection = "auto",
        config: GeminiLiveConfig | None = None,
        **extra: Any,
    ) -> AsyncIterator[LMMSessionOutput]:
        live_config: GeminiLiveConfig = config or ctx.state(GeminiLiveConfig)
        with ctx.scope("gemini_live", live_config):
            ctx.record(
                attributes={
                    "lmm.provider": "gemini",
                    "lmm.model": live_config.model,
                    "lmm.temperature": live_config.temperature,
                    "lmm.max_tokens": live_config.max_tokens,
                    "lmm.seed": live_config.seed,
                    "lmm.tools": [tool["name"] for tool in tools] if tools else [],
                    "lmm.tool_selection": f"{tool_selection}",
                    "lmm.output": f"{output}",
                    "lmm.context": [element.to_str() for element in initial_context or []],
                }
            )

            context_elements: LMMContext
            match initial_context:
                case None:
                    context_elements = ()

                case [*elements]:
                    context_elements = elements

            initial_content: list[Content | ContentDict] = list(
                chain.from_iterable(
                    [context_element_as_content(element) for element in context_elements]
                )
            )

            response_schema: SchemaDict | None
            response_mime_type: str | None
            response_modalities: list[Modality] | None
            response_schema, response_modalities, response_mime_type, output_decoder = (
                output_as_response_declaration(output)
            )

            functions: list[FunctionDeclarationDict] | None
            function_calling_mode: FunctionCallingConfigMode | None
            functions, function_calling_mode = tools_as_tools_config(
                tools,
                tool_selection=tool_selection,
            )

            output_queue = AsyncQueue[LMMSessionOutput]()

            ctx.spawn(
                self._drive_connection,
                session_connection=self._client.aio.live.connect(
                    model=live_config.model,
                    config={
                        "system_instruction": {
                            "parts": [{"text": Instruction.formatted(instruction)}]
                        },
                        "temperature": unwrap_missing(live_config.temperature),
                        "top_p": unwrap_missing(live_config.top_p),
                        "top_k": unwrap_missing(live_config.top_k),
                        "max_output_tokens": unwrap_missing(live_config.max_tokens),
                        "tools": [{"function_declarations": functions}] if functions else None,
                        "response_modalities": response_modalities
                        if response_modalities is not None
                        else None,
                        "seed": unwrap_missing(live_config.seed),
                        "generation_config": {
                            "temperature": unwrap_missing(live_config.temperature),
                            "top_p": unwrap_missing(live_config.top_p),
                            "top_k": unwrap_missing(live_config.top_k),
                            "max_output_tokens": unwrap_missing(live_config.max_tokens),
                            "candidate_count": 1,
                            "seed": unwrap_missing(live_config.seed),
                            "stop_sequences": unwrap_missing(
                                live_config.stop_sequences,
                                transform=as_list,
                            ),
                            "response_mime_type": response_mime_type,
                            "response_schema": response_schema,
                        },
                        "speech_config": unwrap_missing(
                            live_config.speech_voice_name,
                            transform=_speech_config,
                        ),
                    },
                ),
                input_content=initial_content,
                input_stream=input_stream,
                output_stream=output_queue,
            )

            return output_queue

    @ctx.scope("connection", task_group=TaskGroup())
    async def _drive_connection(  # noqa: C901, PLR0912
        self,
        session_connection: AbstractAsyncContextManager[AsyncSession],
        input_content: list[Content | ContentDict],
        input_stream: AsyncIterator[LMMStreamInput],
        output_stream: AsyncQueue[LMMSessionOutput],
    ):
        try:
            async with session_connection as session:
                input_task: Task[None] = ctx.spawn(
                    self._handle_input,
                    session=session,
                    input_stream=input_stream,
                    output_stream=output_stream,
                )
                try:
                    if input_content:
                        await session.send_client_content(
                            turns=input_content,
                            turn_complete=True,
                        )

                    del input_content

                    while not output_stream.is_finished:
                        async for element in session.receive():
                            if tool_calls := element.tool_call:
                                for call in tool_calls.function_calls or ():
                                    if call.name is None or call.id is None:
                                        continue  # skip unknown or incomplete tools

                                    output_stream.enqueue(
                                        LMMToolRequest(
                                            identifier=call.id,
                                            tool=call.name,
                                            arguments=call.args or {},
                                        )
                                    )

                            if content := element.server_content:
                                eod: bool = (
                                    (content.turn_complete or False)
                                    or (content.interrupted or False)
                                    or (content.generation_complete or False)
                                )
                                if generation := content.model_turn:
                                    output_stream.enqueue(
                                        LMMStreamChunk.of(
                                            MultimodalContent.of(
                                                *chain.from_iterable(
                                                    result_part_as_content_or_call(part)
                                                    for part in generation.parts or []
                                                )
                                            ),
                                            eod=eod,
                                        )
                                    )

                                if content.interrupted or False:
                                    output_stream.enqueue(LMMSessionEvent.of("interrupted"))

                                elif eod:
                                    output_stream.enqueue(LMMSessionEvent.of("completed"))
                finally:
                    input_task.cancel()

        except BaseException as exc:
            output_stream.finish(exc)

        else:
            output_stream.finish()

    async def _handle_output(  # noqa: C901
        self,
        session: AsyncSession,
        output_stream: AsyncQueue[LMMSessionOutput],
    ) -> None:
        try:
            while not output_stream.is_finished:
                async for element in session.receive():
                    if tool_calls := element.tool_call:
                        for call in tool_calls.function_calls or ():
                            if call.name is None or call.id is None:
                                continue  # skip unknown or incomplete tools

                            output_stream.enqueue(
                                LMMToolRequest(
                                    identifier=call.id,
                                    tool=call.name,
                                    arguments=call.args or {},
                                )
                            )

                    if content := element.server_content:
                        eod: bool = (
                            (content.turn_complete or False)
                            or (content.interrupted or False)
                            or (content.generation_complete or False)
                        )
                        if generation := content.model_turn:
                            output_stream.enqueue(
                                LMMStreamChunk.of(
                                    MultimodalContent.of(
                                        *chain.from_iterable(
                                            result_part_as_content_or_call(part)
                                            for part in generation.parts or []
                                        )
                                    ),
                                    eod=eod,
                                )
                            )

                        if content.interrupted or False:
                            output_stream.enqueue(LMMSessionEvent.of("interrupted"))

                        elif eod:
                            output_stream.enqueue(LMMSessionEvent.of("completed"))

        except BaseException as exc:
            output_stream.finish(exception=exc)

        else:
            output_stream.finish()

    async def _handle_input(  # noqa: C901
        self,
        session: AsyncSession,
        input_stream: AsyncIterator[LMMStreamInput],
        output_stream: AsyncQueue[LMMSessionOutput],
    ) -> None:
        try:
            async for element in input_stream:
                match element:
                    case LMMStreamChunk() as chunk:
                        for part in chunk.content.parts:
                            match part:
                                case TextContent() as text:
                                    await session.send_realtime_input(
                                        media={
                                            "data": text.text.encode(),
                                            "mime_type": "text/plain",
                                        },
                                    )

                                case MediaData() as media_data:
                                    await session.send_realtime_input(
                                        media={
                                            "data": media_data.data,
                                            "mime_type": media_data.media,
                                        },
                                    )

                                case MediaReference():
                                    raise ValueError("Media reference is not supported")

                                case DataModel() as data:
                                    await session.send_realtime_input(
                                        media={
                                            "data": data.to_json().encode(),
                                            "mime_type": "application/json",
                                        },
                                    )

                    case LMMToolResponse() as tool_response:
                        await session.send_tool_response(
                            function_responses={
                                "id": tool_response.identifier,
                                "name": tool_response.tool,
                                "response": {
                                    "error": [
                                        content_element_as_part(element)
                                        for element in tool_response.content.parts
                                    ],
                                }
                                if tool_response.handling == "error"
                                else {
                                    "output": [
                                        content_element_as_part(element)
                                        for element in tool_response.content.parts
                                    ],
                                },
                            }
                        )

        except BaseException as exc:
            output_stream.finish(exception=exc)

        else:
            output_stream.finish()


def _speech_config(voice_name: str) -> SpeechConfigDict:
    return {
        "voice_config": {
            "prebuilt_voice_config": {"voice_name": voice_name},
        }
    }



from draive import MediaReference, MultimodalContent, MultimodalTagElement


def test_returns_unchanged_with_empty():
    assert (
        MultimodalTagElement.replace(
            "test",
            content=MultimodalContent.empty,
            replacement="replaced",
        )
        == MultimodalContent.empty
    )

    assert (
        MultimodalTagElement.replace(
            "test",
            content=MultimodalContent.empty,
            replacement="replaced",
            strip_tags=True,
        )
        == MultimodalContent.empty
    )


def test_returns_unchanged_without_tag():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("Lorem ipsum"),
        replacement="replaced",
    ) == MultimodalContent.of("Lorem ipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("Lorem ipsum"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("Lorem ipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "Lorem",
        MediaReference.of("http://image", media="image/png"),
        "ipsum",
    )


def test_returns_unchanged_with_other_tag():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<other>Lorem ipsum</other>"),
        replacement="replaced",
    ) == MultimodalContent.of("<other>Lorem ipsum</other>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of('<other key="value">Lorem ipsum</other>'),
        replacement="replaced",
    ) == MultimodalContent.of('<other key="value">Lorem ipsum</other>')

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<other>Lorem ipsum</other>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("<other>Lorem ipsum</other>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<other>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</other>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<other>Lorem",
        MediaReference.of("http://image", media="image/png"),
        "ipsum</other>",
    )


def test_returns_unchanged_without_closing_tag():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum"),
        replacement="replaced",
    ) == MultimodalContent.of("<test>Lorem ipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("<test>Lorem ipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<test>Lorem",
        MediaReference.of("http://image", media="image/png"),
        "ipsum",
    )


def test_returns_unchanged_with_only_closing_tag():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("Lorem </test> ipsum"),
        replacement="replaced",
    ) == MultimodalContent.of("Lorem </test> ipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("Lorem </test> ipsum"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("Lorem </test> ipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "Lorem </test> ",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "Lorem </test> ",
        MediaReference.of("http://image", media="image/png"),
        "ipsum",
    )


def test_returns_unchanged_with_reversed_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("</test>Lorem <test> ipsum"),
        replacement="replaced",
    ) == MultimodalContent.of("</test>Lorem <test> ipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("</test>Lorem <test> ipsum"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("</test>Lorem <test> ipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "</test>Lorem <test> ",
            MediaReference.of("http://image", media="image/png"),
            "ipsum",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "</test>Lorem <test> ",
        MediaReference.of("http://image", media="image/png"),
        "ipsum",
    )


def test_returns_unchanged_with_malformed_opening_tag():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<testx>Lorem ipsum</test>"),
        replacement="replaced",
    ) == MultimodalContent.of("<testx>Lorem ipsum</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test key=invalid>Lorem ipsum</test>"),
        replacement="replaced",
    ) == MultimodalContent.of("<test key=invalid>Lorem ipsum</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<testx>Lorem ipsum</test>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("<testx>Lorem ipsum</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<testx>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<testx>Lorem",
        MediaReference.of("http://image", media="image/png"),
        "ipsum</test>",
    )


def test_returns_unchanged_with_malformed_closing_tag():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</testx>"),
        replacement="replaced",
    ) == MultimodalContent.of("<test>Lorem ipsum</testx>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of('<test key="value">Lorem ipsum</testx>'),
        replacement="replaced",
    ) == MultimodalContent.of('<test key="value">Lorem ipsum</testx>')

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</testx>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("<test>Lorem ipsum</testx>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</testx>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<test>Lorem",
        MediaReference.of("http://image", media="image/png"),
        "ipsum</testx>",
    )


def test_returns_replaced_with_valid_tag():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</test>"),
        replacement="replaced",
    ) == MultimodalContent.of("<test>replaced</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</test>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("replaced")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<test>replaced</test>",
    )


def test_returns_replaced_outer_with_multiple_nested_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Other<test>Lorem ipsum</test></test>"),
        replacement="replaced",
    ) == MultimodalContent.of("<test>replaced</test></test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Other<test>Lorem ipsum</test></test>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("replaced</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<test>Other<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test></test>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<test>replaced</test></test>",
    )


def test_returns_replaced_with_surrounded_tag():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("Lorem<test>Lorem ipsum</test>ipsum"),
        replacement="replaced",
    ) == MultimodalContent.of("Lorem<test>replaced</test>ipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("Lorem<test>Lorem ipsum</test>ipsum"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("Loremreplacedipsum")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            MediaReference.of("http://image", media="image/png"),
            "Lorem<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>ipsum",
            MediaReference.of("http://image", media="image/png"),
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        MediaReference.of("http://image", media="image/png"),
        "Lorem<test>replaced</test>ipsum",
        MediaReference.of("http://image", media="image/png"),
    )


def test_returns_unchanged_with_nested_in_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<other>Other<test>Lorem ipsum</test></other>"),
        replacement="replaced",
    ) == MultimodalContent.of("<other>Other<test>Lorem ipsum</test></other>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<other>Other<test>Lorem ipsum</test></other>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("<other>Other<test>Lorem ipsum</test></other>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<other>Other<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test></other>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<other>Other<test>Lorem",
        MediaReference.of("http://image", media="image/png"),
        "ipsum</test></other>",
    )


def test_returns_replaced_content_with_fake_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem<ipsum</test>"),
        replacement="replaced",
    ) == MultimodalContent.of("<test>replaced</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<te<test>Lorem<ipsum</test>"),
        replacement="replaced",
    ) == MultimodalContent.of("<te<test>replaced</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem<ipsum</test>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("replaced")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<te<test>Lorem<ipsum</test>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("<tereplaced")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<test>Lorem<",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<test>replaced</test>",
    )

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<te<test>Lorem<",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<te<test>replaced</test>",
    )


def test_returns_replaced_with_other_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<other>Other</other><test>Lorem</more>ipsum</test><more>"),
        replacement="replaced",
    ) == MultimodalContent.of("<other>Other</other><test>replaced</test><more>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<other>Other</other><test>Lorem</more>ipsum</test><more>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("<other>Other</other>replaced<more>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<other>Other</other><test>Lorem</more>",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test><more>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<other>Other</other><test>replaced</test><more>",
    )


def test_returns_first_replaced_with_multiple_valid_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</test><test>Lorem ipsum</test>"),
        replacement="replaced",
    ) == MultimodalContent.of("<test>replaced</test><test>Lorem ipsum</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</test><test>Lorem ipsum</test>"),
        replacement="replaced",
        strip_tags=True,
    ) == MultimodalContent.of("replaced<test>Lorem ipsum</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
            "<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
        ),
        replacement="replaced",
    ) == MultimodalContent.of(
        "<test>replaced</test>",
        "<test>Lorem",
        MediaReference.of("http://image", media="image/png"),
        "ipsum</test>",
    )


def test_returns_all_replaced_with_multiple_valid_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</test><test>Lorem ipsum</test>"),
        replacement="replaced",
        replace_all=True,
    ) == MultimodalContent.of("<test>replaced</test><test>replaced</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>Lorem ipsum</test><test>Lorem ipsum</test>"),
        replacement="replaced",
        strip_tags=True,
        replace_all=True,
    ) == MultimodalContent.of("replacedreplaced")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
            "<test>Lorem",
            MediaReference.of("http://image", media="image/png"),
            "ipsum</test>",
        ),
        replacement="replaced",
        replace_all=True,
    ) == MultimodalContent.of(
        "<test>replaced</test><test>replaced</test>",
    )


### generated ###


def test_handles_empty_replacement():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>content</test>"),
        replacement="",
    ) == MultimodalContent.of("<test></test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>content</test>"),
        replacement="",
        strip_tags=True,
    ) == MultimodalContent.of("")


def test_handles_whitespace_replacement():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>content</test>"),
        replacement="   \n\t  ",
    ) == MultimodalContent.of("<test>   \n\t  </test>")


def test_handles_self_closing_tag_replacement():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test/>"),
        replacement="content",
    ) == MultimodalContent.of("<test>content</test>")

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test/>"),
        replacement="content",
        strip_tags=True,
    ) == MultimodalContent.of("content")


def test_handles_multiple_self_closing_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test/><other/><test/>"),
        replacement="content",
        replace_all=True,
    ) == MultimodalContent.of("<test>content</test><other/><test>content</test>")


def test_preserves_tag_attributes_on_replacement():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of('<test id="1" class="example">old</test>'),
        replacement="new",
    ) == MultimodalContent.of('<test id="1" class="example">new</test>')


def test_handles_nested_tags_with_same_name():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test><test><test>content</test></test></test>"),
        replacement="replaced",
        replace_all=True,
    ) == MultimodalContent.of("<test>replaced</test></test></test>")


def test_handles_mixed_content_replacement():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<test>",
            MediaReference.of("http://image1", media="image/png"),
            "text",
            MediaReference.of("http://image2", media="image/png"),
            "</test>",
        ),
        replacement=MediaReference.of("http://replacement", media="image/jpeg"),
    ) == MultimodalContent.of(
        "<test>", MediaReference.of("http://replacement", media="image/jpeg"), "</test>"
    )


def test_handles_escaped_characters_in_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of('<test escaped="\\n\\t\\"">content</test>'),
        replacement="replaced",
    ) == MultimodalContent.of('<test escaped="\\n\\t\\"">replaced</test>')


def test_handles_adjacent_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>one</test><test>two</test><test>three</test>"),
        replacement="replaced",
        replace_all=True,
    ) == MultimodalContent.of("<test>replaced</test><test>replaced</test><test>replaced</test>")


def test_handles_tags_with_numeric_suffixes():
    assert MultimodalTagElement.replace(
        "test1",
        content=MultimodalContent.of("<test1>content</test1>"),
        replacement="replaced",
    ) == MultimodalContent.of("<test1>replaced</test1>")


def test_handles_tags_with_special_characters():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>< > & \" '</test>"),
        replacement="replaced",
    ) == MultimodalContent.of("<test>replaced</test>")


def test_handles_replacement_with_tags():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>content</test>"),
        replacement="<other>nested</other>",
    ) == MultimodalContent.of("<test><other>nested</other></test>")


def test_handles_multiline_content():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>\nline1\nline2\n</test>"),
        replacement="replaced",
    ) == MultimodalContent.of("<test>replaced</test>")


def test_handles_mixed_case_sensitivity():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<TEST>content</TEST>"),
        replacement="replaced",
    ) == MultimodalContent.of("<TEST>content</TEST>")


def test_handles_multiple_attributes_with_quotes():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of('<test a="1" b="2" c="3">content</test>'),
        replacement="replaced",
    ) == MultimodalContent.of('<test a="1" b="2" c="3">replaced</test>')


def test_handles_complex_nested_structure():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<outer><test>one</test><middle><test>two</test></middle><test>three</test></outer>"
        ),
        replacement="replaced",
        replace_all=True,
    ) == MultimodalContent.of(
        "<outer><test>one</test><middle><test>two</test></middle><test>three</test></outer>"
    )

    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of(
            "<outer><test>one</test><middle></outer><test>two</test></middle><test>three</test>"
        ),
        replacement="replaced",
        replace_all=True,
    ) == MultimodalContent.of(
        "<outer><test>one</test><middle></outer><test>replaced</test></middle><test>replaced</test>"
    )


def test_handles_replacement_with_multiple_media():
    assert MultimodalTagElement.replace(
        "test",
        content=MultimodalContent.of("<test>old</test>"),
        replacement=MultimodalContent.of(
            MediaReference.of("http://image1", media="image/png"),
            "text",
            MediaReference.of("http://image2", media="image/jpeg"),
        ),
    ) == MultimodalContent.of(
        "<test>",
        MediaReference.of("http://image1", media="image/png"),
        "text",
        MediaReference.of("http://image2", media="image/jpeg"),
        "</test>",
    )



from draive.tools.function import FunctionTool, tool
from draive.tools.state import Tools
from draive.tools.toolbox import Toolbox
from draive.tools.types import (
    Tool,
    ToolAvailabilityChecking,
    ToolErrorFormatting,
    ToolExcepion,
    ToolHandling,
    ToolResultFormatting,
    ToolsFetching,
)

__all__ = (
    "FunctionTool",
    "Tool",
    "ToolAvailabilityChecking",
    "ToolErrorFormatting",
    "ToolExcepion",
    "ToolHandling",
    "ToolResultFormatting",
    "Toolbox",
    "Tools",
    "ToolsFetching",
    "tool",
)



from typing import Any, Protocol, runtime_checkable

from haiway import State

from draive.commons import META_EMPTY, Meta
from draive.multimodal import MultimodalContent

__all__ = (
    "GuardrailsAnonymizedContent",
    "GuardrailsContentAnonymizing",
    "GuardrailsContentDeanonymizing",
)


@runtime_checkable
class GuardrailsContentDeanonymizing(Protocol):
    async def __call__(
        self,
        content: MultimodalContent,
        /,
    ) -> MultimodalContent: ...


class GuardrailsAnonymizedContent(State):
    content: MultimodalContent
    deanonymize: GuardrailsContentDeanonymizing | None = None
    meta: Meta = META_EMPTY


@runtime_checkable
class GuardrailsContentAnonymizing(Protocol):
    async def __call__(
        self,
        content: MultimodalContent,
        /,
        **extra: Any,
    ) -> GuardrailsAnonymizedContent: ...



from collections.abc import Callable, Mapping, Sequence
from datetime import date, datetime, time
from enum import IntEnum, StrEnum
from types import EllipsisType, NoneType, UnionType
from typing import (
    Any,
    Final,
    Literal,
    NotRequired,
    Required,
    TypedDict,
    Union,
    cast,
    final,
    is_typeddict,
)
from uuid import UUID

from haiway import Missing
from haiway.state import AttributeAnnotation

from draive.commons import Meta
from draive.parameters.types import ParameterValidation, ParameterValidationContext
from draive.parameters.validation import ParameterValidator

__all__ = (
    "ParameterSpecification",
    "ParametersSpecification",
    "parameter_specification",
    "validated_specification",
)


@final
class ParameterNoneSpecification(TypedDict, total=False):
    type: Required[Literal["null"]]
    description: NotRequired[str]


@final
class ParameterBoolSpecification(TypedDict, total=False):
    type: Required[Literal["boolean"]]
    description: NotRequired[str]


@final
class ParameterIntegerSpecification(TypedDict, total=False):
    type: Required[Literal["integer"]]
    description: NotRequired[str]


@final
class ParameterNumberSpecification(TypedDict, total=False):
    type: Required[Literal["number"]]
    description: NotRequired[str]


@final
class ParameterStringSpecification(TypedDict, total=False):
    type: Required[Literal["string"]]
    description: NotRequired[str]
    format: NotRequired[
        Literal[
            "uri",
            "uuid",
            "date",
            "time",
            "date-time",
        ]
    ]


@final
class ParameterStringEnumSpecification(TypedDict, total=False):
    type: Required[Literal["string"]]
    enum: Required[Sequence[str]]
    description: NotRequired[str]


@final
class ParameterIntegerEnumSpecification(TypedDict, total=False):
    type: Required[Literal["integer"]]
    enum: Required[Sequence[int]]
    description: NotRequired[str]


@final
class ParameterNumberEnumSpecification(TypedDict, total=False):
    type: Required[Literal["number"]]
    enum: Required[Sequence[float]]
    description: NotRequired[str]


@final
class ParameterUnionSpecification(TypedDict, total=False):
    oneOf: Required[Sequence["ParameterSpecification"]]
    description: NotRequired[str]


@final
class ParameterArraySpecification(TypedDict, total=False):
    type: Required[Literal["array"]]
    items: NotRequired["ParameterSpecification"]
    description: NotRequired[str]


@final
class ParameterTupleSpecification(TypedDict, total=False):
    type: Required[Literal["array"]]
    prefixItems: Required[Sequence["ParameterSpecification"]]
    description: NotRequired[str]


@final
class ParameterDictSpecification(TypedDict, total=False):
    type: Required[Literal["object"]]
    additionalProperties: Required["ParameterSpecification"]
    description: NotRequired[str]
    required: NotRequired[Sequence[str]]


@final
class ParameterObjectSpecification(TypedDict, total=False):
    type: Required[Literal["object"]]
    properties: Required[Mapping[str, "ParameterSpecification"]]
    title: NotRequired[str]
    description: NotRequired[str]
    required: NotRequired[Sequence[str]]


@final
class ParameterAnyObjectSpecification(TypedDict, total=False):
    type: Required[Literal["object"]]
    additionalProperties: Required[Literal[True]]
    description: NotRequired[str]


ReferenceParameterSpecification = TypedDict(
    "ReferenceParameterSpecification",
    {
        "$ref": Required[str],
        "description": NotRequired[str],
    },
    total=False,
)


type ParameterSpecification = (
    ParameterUnionSpecification
    | ParameterNoneSpecification
    | ParameterStringEnumSpecification
    | ParameterStringSpecification
    | ParameterIntegerEnumSpecification
    | ParameterIntegerSpecification
    | ParameterNumberEnumSpecification
    | ParameterNumberSpecification
    | ParameterBoolSpecification
    | ParameterTupleSpecification
    | ParameterArraySpecification
    | ParameterObjectSpecification
    | ParameterDictSpecification
    | ParameterAnyObjectSpecification
    | ReferenceParameterSpecification
)

type ParametersSpecification = ParameterObjectSpecification


def validated_specification(
    specification: dict[str, Any],
    /,
) -> ParameterObjectSpecification:
    with ParameterValidationContext().scope("specification") as validation_context:
        return _specification_validation(
            specification,
            context=validation_context,
        )


_specification_validation: Final[ParameterValidation[ParameterObjectSpecification]] = (
    ParameterValidator.of_typed_dict(ParameterObjectSpecification)
)


def parameter_specification(  # noqa: PLR0911
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if specification := SPECIFICATIONS.get(annotation.origin):
        return specification(annotation, description)

    elif hasattr(annotation.origin, "__PARAMETERS_SPECIFICATION__"):
        if description := description:
            return cast(
                ParameterSpecification,
                {
                    **annotation.origin.__PARAMETERS_SPECIFICATION__,
                    "description": description,
                },
            )

        else:
            return annotation.origin.__PARAMETERS_SPECIFICATION__

    elif is_typeddict(annotation.origin):
        return _prepare_specification_of_typed_dict(
            annotation,
            description=description,
        )

    elif issubclass(annotation.origin, IntEnum):
        if description := description:
            return {
                "type": "integer",
                "enum": list(annotation.origin),
                "description": description,
            }

        else:
            return {
                "type": "integer",
                "enum": list(annotation.origin),
            }

    elif issubclass(annotation.origin, StrEnum):
        if description := description:
            return {
                "type": "string",
                "enum": list(annotation.origin),
                "description": description,
            }

        else:
            return {
                "type": "string",
                "enum": list(annotation.origin),
            }

    else:
        raise TypeError(f"Unsupported type annotation: {annotation}")


def _prepare_specification_of_any(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "object",
            "additionalProperties": True,
            "description": description,
        }

    else:
        return {
            "type": "object",
            "additionalProperties": True,
        }


def _prepare_specification_of_none(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "null",
            "description": description,
        }

    else:
        return {
            "type": "null",
        }


def _prepare_specification_of_missing(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "null",
            "description": description,
        }

    else:
        return {
            "type": "null",
        }


def _prepare_specification_of_literal(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if all(isinstance(element, str) for element in annotation.arguments):
        if description := description:
            return {
                "type": "string",
                "enum": annotation.arguments,
                "description": description,
            }

        else:
            return {
                "type": "string",
                "enum": annotation.arguments,
            }
    else:
        raise TypeError(f"Unsupported type annotation: {annotation}")


def _prepare_specification_of_sequence(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "array",
            "items": parameter_specification(
                annotation.arguments[0],
                description=None,
            ),
            "description": description,
        }

    else:
        return {
            "type": "array",
            "items": parameter_specification(
                annotation.arguments[0],
                description=None,
            ),
        }


def _prepare_specification_of_mapping(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "object",
            "additionalProperties": parameter_specification(
                annotation.arguments[1],
                description=None,
            ),
            "description": description,
        }

    else:
        return {
            "type": "object",
            "additionalProperties": parameter_specification(
                annotation.arguments[1],
                description=None,
            ),
        }


def _prepare_specification_of_meta(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    # TODO: prepare actual property types declaration
    if description := description:
        return {
            "type": "object",
            "additionalProperties": True,
            "description": description,
        }

    else:
        return {
            "type": "object",
            "additionalProperties": True,
        }


def _prepare_specification_of_tuple(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if (
        annotation.arguments[-1].origin == Ellipsis
        or annotation.arguments[-1].origin == EllipsisType
    ):
        if description := description:
            return {
                "type": "array",
                "items": parameter_specification(annotation.arguments[0], description=None),
                "description": description,
            }

        else:
            return {
                "type": "array",
                "items": parameter_specification(annotation.arguments[0], description=None),
            }

    elif description := description:
        return {
            "type": "array",
            "prefixItems": [
                parameter_specification(element, description=None)
                for element in annotation.arguments
            ],
            "description": description,
        }

    else:
        return {
            "type": "array",
            "prefixItems": [
                parameter_specification(element, description=None)
                for element in annotation.arguments
            ],
        }


def _prepare_specification_of_union(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "oneOf": [
                parameter_specification(cast(AttributeAnnotation, argument), description=None)
                for argument in annotation.arguments
            ],
            "description": description,
        }

    else:
        return {
            "oneOf": [
                parameter_specification(cast(AttributeAnnotation, argument), description=None)
                for argument in annotation.arguments
            ],
        }


def _prepare_specification_of_bool(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "boolean",
            "description": description,
        }

    else:
        return {
            "type": "boolean",
        }


def _prepare_specification_of_int(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "integer",
            "description": description,
        }

    else:
        return {
            "type": "integer",
        }


def _prepare_specification_of_float(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "number",
            "description": description,
        }

    else:
        return {
            "type": "number",
        }


def _prepare_specification_of_str(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "description": description,
        }

    else:
        return {
            "type": "string",
        }


def _prepare_specification_of_uuid(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "format": "uuid",
            "description": description,
        }

    else:
        return {
            "type": "string",
            "format": "uuid",
        }


def _prepare_specification_of_date(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "format": "date",
            "description": description,
        }

    else:
        return {
            "type": "string",
            "format": "date",
        }


def _prepare_specification_of_datetime(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "format": "date-time",
            "description": description,
        }

    else:
        return {
            "type": "string",
            "format": "date-time",
        }


def _prepare_specification_of_time(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    if description := description:
        return {
            "type": "string",
            "format": "time",
            "description": description,
        }

    else:
        return {
            "type": "string",
            "format": "time",
        }


def _prepare_specification_of_typed_dict(
    annotation: AttributeAnnotation,
    /,
    description: str | None,
) -> ParameterSpecification:
    required: list[str] = []
    properties: dict[str, ParameterSpecification] = {}

    for key, element in annotation.extra["attributes"].items():
        properties[key] = parameter_specification(
            element,
            description=None,
        )

        if getattr(element, "required", True):
            required.append(key)

    if description := description:
        return {
            "type": "object",
            "properties": properties,
            "required": required,
            "description": description,
        }

    else:
        return {
            "type": "object",
            "properties": properties,
            "required": required,
        }


SPECIFICATIONS: Mapping[
    Any, Callable[[AttributeAnnotation, str | None], ParameterSpecification]
] = {
    Any: _prepare_specification_of_any,
    NoneType: _prepare_specification_of_none,
    Missing: _prepare_specification_of_missing,
    bool: _prepare_specification_of_bool,
    int: _prepare_specification_of_int,
    float: _prepare_specification_of_float,
    str: _prepare_specification_of_str,
    tuple: _prepare_specification_of_tuple,
    Literal: _prepare_specification_of_literal,
    Sequence: _prepare_specification_of_sequence,
    Mapping: _prepare_specification_of_mapping,
    Meta: _prepare_specification_of_meta,
    UUID: _prepare_specification_of_uuid,
    date: _prepare_specification_of_date,
    datetime: _prepare_specification_of_datetime,
    time: _prepare_specification_of_time,
    Union: _prepare_specification_of_union,
    UnionType: _prepare_specification_of_union,
}



import json
from collections.abc import Iterator, Mapping, Sequence
from types import EllipsisType
from typing import Any, ClassVar, Final, Self, cast, final

__all__ = (
    "META_EMPTY",
    "Meta",
    "MetaTags",
    "MetaValue",
    "MetaValues",
    "validated_meta_value",
)

type MetaValue = Mapping[str, MetaValue] | Sequence[MetaValue] | str | float | int | bool | None
type MetaValues = Mapping[str, MetaValue]
type MetaTags = Sequence[str]


@final
class Meta(Mapping[str, MetaValue]):
    _: ClassVar[Self]
    __IMMUTABLE__: ClassVar[EllipsisType] = ...

    __slots__ = ("_values",)

    def __init__(
        self,
        values: MetaValues,
        /,
    ):
        self._values: MetaValues
        if isinstance(values, Meta):  # avoid wrapping twice
            object.__setattr__(
                self,
                "_values",
                values._values,
            )

        else:
            object.__setattr__(
                self,
                "_values",
                values,
            )

    @classmethod
    def of(
        cls,
        meta: Self | MetaValues | None,
    ) -> Self:
        match meta:
            case None:
                return cast(Self, META_EMPTY)

            case Meta():
                return cast(Self, meta)

            case mapping:
                return cls({key: validated_meta_value(value) for key, value in mapping.items()})

    @classmethod
    def from_mapping(
        cls,
        mapping: Mapping[str, Any],
        /,
    ) -> Self:
        return cls({key: validated_meta_value(value) for key, value in mapping.items()})

    @classmethod
    def from_json(
        cls,
        value: str | bytes,
        /,
    ) -> Self:
        match json.loads(value):
            case {**values}:
                return cls({key: validated_meta_value(val) for key, val in values.items()})

            case other:
                raise ValueError(f"Invalid json: {other}")

    def to_str(self) -> str:
        return self.__str__()

    def to_mapping(
        self,
    ) -> Mapping[str, Any]:
        return self._values

    def to_json(
        self,
    ) -> str:
        return json.dumps(self._values)

    @property
    def kind(self) -> str | None:
        match self._values.get("kind"):
            case str() as kind:
                return kind

            case _:
                return None

    def with_kind(
        self,
        kind: str,
        /,
    ) -> Self:
        return self.__class__(
            {
                **self._values,
                "kind": kind,
            }
        )

    @property
    def name(self) -> str | None:
        match self._values.get("name"):
            case str() as name:
                return name

            case _:
                return None

    def with_name(
        self,
        name: str,
        /,
    ) -> Self:
        return self.__class__(
            {
                **self._values,
                "name": name,
            }
        )

    @property
    def description(self) -> str | None:
        match self._values.get("description"):
            case str() as description:
                return description

            case _:
                return None

    def with_description(
        self,
        description: str,
        /,
    ) -> Self:
        return self.__class__(
            {
                **self._values,
                "description": description,
            }
        )

    @property
    def tags(self) -> MetaTags:
        match self._values.get("tags"):
            case [*tags]:
                return tuple(tag for tag in tags if isinstance(tag, str))

            case _:
                return ()

    def with_tags(
        self,
        tags: MetaTags,
        /,
    ) -> Self:
        match self._values.get("tags"):
            case [*current_tags]:
                return self.__class__(
                    {
                        **self._values,
                        "tags": (
                            *current_tags,
                            *(validated_meta_value(tag) for tag in tags if tag not in current_tags),
                        ),
                    }
                )

            case _:
                return self.__class__({**self._values, "tags": tags})

    def has_tags(
        self,
        tags: MetaTags,
        /,
    ) -> bool:
        match self._values.get("tags"):
            case [*meta_tags]:
                return all(tag in meta_tags for tag in tags)

            case _:
                return False

    def merged_with(
        self,
        values: Self | MetaValues,
        /,
    ) -> Self:
        if not values:
            return self  # do not make a copy when nothing will be updated

        return self.__class__(
            {
                **self._values,  # already validated
                **{key: validated_meta_value(value) for key, value in values.items()},
            }
        )

    def updated(
        self,
        **values: MetaValue,
    ) -> Self:
        return self.__replace__(**values)

    def __replace__(
        self,
        **values: Any,
    ) -> Self:
        return self.merged_with(values)

    def __bool__(self) -> bool:
        return bool(self._values)

    def __contains__(
        self,
        element: Any,
    ) -> bool:
        return element in self._values

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def __setitem__(
        self,
        key: str,
        value: Any,
    ) -> MetaValue:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" item - '{key}' cannot be modified"
        )

    def __delitem__(
        self,
        key: str,
    ) -> MetaValue:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" item - '{key}' cannot be deleted"
        )

    def __getitem__(
        self,
        key: str,
    ) -> MetaValue:
        return self._values[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._values)

    def __len__(self) -> int:
        return len(self._values)

    def __copy__(self) -> Self:
        return self  # Metadata is immutable, no need to provide an actual copy

    def __deepcopy__(
        self,
        memo: dict[int, Any] | None,
    ) -> Self:
        return self  # Metadata is immutable, no need to provide an actual copy


def validated_meta_value(value: Any) -> MetaValue:
    match value:
        case None:
            return value

        case str():
            return value

        case int():
            return value

        case float():
            return value

        case [*values]:
            return tuple(validated_meta_value(value) for value in values)

        case {**values}:
            return {key: validated_meta_value(value) for key, value in values.items()}

        case other:
            raise TypeError(f"Invalid Meta value: {type(other)}")


META_EMPTY: Final[Meta] = Meta({})



import json
from collections.abc import AsyncGenerator, AsyncIterator, Callable, Iterable
from itertools import chain
from typing import Any, Literal, cast, overload
from uuid import uuid4

from haiway import ObservabilityLevel, as_list, ctx
from mistralai import (
    CompletionEvent,
    CompletionResponseStreamChoice,
    DeltaMessage,
    ToolCall,
    ToolTypedDict,
)
from mistralai.models import (
    AssistantMessage,
    ChatCompletionChoice,
    ChatCompletionRequestToolChoiceTypedDict,
    ChatCompletionResponse,
    MessagesTypedDict,
    ResponseFormatTypedDict,
    StopTypedDict,
)
from mistralai.types.basemodel import Unset
from mistralai.utils.eventstreaming import EventStreamAsync

from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContext,
    LMMOutput,
    LMMOutputSelection,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.lmm.types import LMMStreamChunk, LMMStreamOutput
from draive.mistral.api import MistralAPI
from draive.mistral.config import MistralChatConfig
from draive.mistral.lmm import (
    content_chunk_as_content_element,
    content_element_as_content_chunk,
    context_element_as_messages,
    output_as_response_declaration,
    tools_as_tool_config,
)
from draive.mistral.types import MistralException
from draive.mistral.utils import unwrap_missing_to_none, unwrap_missing_to_unset
from draive.multimodal import Multimodal, MultimodalContent

__all__ = ("MistralLMMGeneration",)


class MistralLMMGeneration(MistralAPI):
    def lmm(self) -> LMM:
        return LMM(completing=self.lmm_completion)

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        prefill: Multimodal | None = None,
        config: MistralChatConfig | None = None,
        output: LMMOutputSelection,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        prefill: Multimodal | None = None,
        config: MistralChatConfig | None = None,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        prefill: Multimodal | None = None,
        config: MistralChatConfig | None = None,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput:
        completion_config: MistralChatConfig = config or ctx.state(MistralChatConfig)
        with ctx.scope("mistral_lmm_completion", completion_config):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "lmm.provider": "mistral",
                    "lmm.model": completion_config.model,
                    "lmm.temperature": completion_config.temperature,
                    "lmm.max_tokens": completion_config.max_tokens,
                    "lmm.seed": completion_config.seed,
                    "lmm.tools": [tool["name"] for tool in tools] if tools else None,
                    "lmm.tool_selection": f"{tool_selection}" if tools else None,
                    "lmm.stream": stream,
                    "lmm.output": f"{output}",
                    "lmm.instruction": f"{instruction}",
                    "lmm.context": [element.to_str() for element in context],
                },
            )

            messages: list[MessagesTypedDict] = list(
                chain.from_iterable([context_element_as_messages(element) for element in context])
            )

            if prefill:
                messages.append(
                    {
                        "role": "assistant",
                        "content": [
                            content_element_as_content_chunk(element)
                            for element in MultimodalContent.of(prefill).parts
                        ],
                        "prefix": True,
                    }
                )

            if instruction:
                messages = [
                    {
                        "role": "system",
                        "content": Instruction.formatted(instruction),
                    },
                    *messages,
                ]

            response_format, output_decoder = output_as_response_declaration(output)

            tool_choice, tools_list = tools_as_tool_config(
                tools,
                tool_selection=tool_selection,
            )

            if stream:
                return await self._completion_stream(
                    model=completion_config.model,
                    messages=messages,
                    temperature=completion_config.temperature,
                    top_p=unwrap_missing_to_none(completion_config.top_p),
                    max_tokens=unwrap_missing_to_unset(completion_config.max_tokens),
                    stop=as_list(unwrap_missing_to_none(completion_config.stop_sequences)),
                    random_seed=unwrap_missing_to_unset(completion_config.seed),
                    response_format=response_format,
                    tools=tools_list,
                    tool_choice=tool_choice,
                    output_decoder=output_decoder,
                )

            else:
                return await self._completion(
                    model=completion_config.model,
                    messages=messages,
                    temperature=completion_config.temperature,
                    top_p=unwrap_missing_to_none(completion_config.top_p),
                    max_tokens=unwrap_missing_to_unset(completion_config.max_tokens),
                    stop=as_list(unwrap_missing_to_none(completion_config.stop_sequences)),
                    random_seed=unwrap_missing_to_unset(completion_config.seed),
                    response_format=response_format,
                    tools=tools_list,
                    tool_choice=tool_choice,
                    output_decoder=output_decoder,
                )

    async def _completion(
        self,
        model: str,
        messages: list[MessagesTypedDict],
        temperature: float,
        top_p: float | None,
        max_tokens: int | Unset,
        stop: StopTypedDict | None,
        random_seed: int | Unset,
        response_format: ResponseFormatTypedDict,
        tools: list[ToolTypedDict] | Unset,
        tool_choice: ChatCompletionRequestToolChoiceTypedDict | None,
        output_decoder: Callable[[MultimodalContent], MultimodalContent],
    ) -> LMMOutput:
        completion: ChatCompletionResponse = await self._client.chat.complete_async(
            model=model,
            messages=messages,
            temperature=temperature,
            top_p=top_p,
            max_tokens=max_tokens,
            stop=stop,
            random_seed=random_seed,
            response_format=response_format,
            tools=tools,
            tool_choice=tool_choice,
            stream=False,
        )

        if usage := completion.usage:
            ctx.record(
                ObservabilityLevel.INFO,
                metric="lmm.input_tokens",
                value=usage.prompt_tokens,
                unit="tokens",
                attributes={"lmm.model": completion.model},
            )
            ctx.record(
                ObservabilityLevel.INFO,
                metric="lmm.output_tokens",
                value=usage.completion_tokens,
                unit="tokens",
                attributes={"lmm.model": completion.model},
            )

        if not completion.choices:
            raise MistralException("Invalid Mistral completion - missing choices!", completion)

        completion_choice: ChatCompletionChoice = completion.choices[0]

        match completion_choice.finish_reason:
            case "stop" | "tool_calls":
                pass  # process results

            case "length":
                raise MistralException(
                    "Invalid Mistral completion - exceeded maximum length!",
                    completion,
                )

            case "error":
                raise MistralException(
                    "Mistral completion generation failed!",
                    completion,
                )

        completion_message: AssistantMessage = completion_choice.message

        lmm_completion: LMMCompletion | None
        if content := completion_message.content:
            match content:
                case str() as string:
                    lmm_completion = LMMCompletion.of(output_decoder(MultimodalContent.of(string)))

                case chunks:
                    lmm_completion = LMMCompletion.of(
                        output_decoder(
                            MultimodalContent.of(
                                *[content_chunk_as_content_element(chunk) for chunk in chunks]
                            )
                        ),
                    )

        else:
            lmm_completion = None

        if tool_calls := completion_message.tool_calls:
            completion_tool_calls = LMMToolRequests(
                content=lmm_completion.content if lmm_completion else None,
                requests=[
                    LMMToolRequest(
                        identifier=call.id or uuid4().hex,
                        tool=call.function.name,
                        arguments=json.loads(call.function.arguments)
                        if isinstance(call.function.arguments, str)
                        else call.function.arguments,
                    )
                    for call in tool_calls
                ],
            )
            ctx.record(
                ObservabilityLevel.INFO,
                event="lmm.tool_requests",
                attributes={"lmm.tools": [call.function.name for call in tool_calls]},
            )
            return completion_tool_calls

        elif lmm_completion:
            ctx.record(
                ObservabilityLevel.INFO,
                event="lmm.completion",
            )
            return lmm_completion

        else:
            raise MistralException("Invalid Mistral completion, missing content!", completion)

    async def _completion_stream(  # noqa: C901, PLR0915
        self,
        model: str,
        messages: list[MessagesTypedDict],
        temperature: float,
        top_p: float | None,
        max_tokens: int | Unset,
        stop: StopTypedDict | None,
        random_seed: int | Unset,
        response_format: ResponseFormatTypedDict,
        tools: list[ToolTypedDict] | Unset,
        tool_choice: ChatCompletionRequestToolChoiceTypedDict | None,
        output_decoder: Callable[[MultimodalContent], MultimodalContent],
    ) -> AsyncIterator[LMMStreamOutput]:
        response_stream: EventStreamAsync[CompletionEvent] = await self._client.chat.stream_async(
            model=model,
            messages=messages,
            temperature=temperature,
            top_p=top_p,
            max_tokens=max_tokens,
            stop=stop,
            random_seed=random_seed,
            response_format=response_format,
            tools=tools,
            tool_choice=tool_choice,
            stream=True,
        )

        async def stream() -> AsyncGenerator[LMMStreamOutput]:  # noqa: C901, PLR0912, PLR0915
            async with response_stream:
                accumulated_tool_calls: list[ToolCall] = []
                async for completion_chunk in response_stream:
                    if usage := completion_chunk.data.usage:
                        ctx.record(
                            ObservabilityLevel.INFO,
                            metric="lmm.input_tokens",
                            value=usage.prompt_tokens,
                            unit="tokens",
                            attributes={"lmm.model": completion_chunk.data.model},
                        )
                        ctx.record(
                            ObservabilityLevel.INFO,
                            metric="lmm.output_tokens",
                            value=usage.completion_tokens,
                            unit="tokens",
                            attributes={"lmm.model": completion_chunk.data.model},
                        )

                    if not completion_chunk.data.choices:
                        raise MistralException(
                            "Invalid Mistral completion - missing choices!",
                            completion_chunk.data,
                        )

                    completion_choice: CompletionResponseStreamChoice = (
                        completion_chunk.data.choices[0]
                    )

                    completion_delta: DeltaMessage = completion_choice.delta
                    if content := completion_delta.content:
                        match content:
                            case str() as string:
                                yield LMMStreamChunk.of(
                                    output_decoder(MultimodalContent.of(string))
                                )

                            case chunks:
                                yield LMMStreamChunk.of(
                                    output_decoder(
                                        MultimodalContent.of(
                                            *[
                                                content_chunk_as_content_element(chunk)
                                                for chunk in chunks
                                            ]
                                        )
                                    )
                                )

                    if tool_calls := completion_delta.tool_calls:
                        if not accumulated_tool_calls:
                            accumulated_tool_calls = sorted(
                                tool_calls,
                                key=lambda call: call.index or 0,
                            )

                        else:
                            for tool_call in tool_calls:
                                assert tool_call.index, "Can't identify function call without index"  # nosec: B101

                                # "null" is a dafault value...
                                if tool_call.id and tool_call.id != "null":
                                    accumulated_tool_calls[tool_call.index].id = tool_call.id

                                if tool_call.function.name:
                                    accumulated_tool_calls[
                                        tool_call.index
                                    ].function.name += tool_call.function.name

                                if isinstance(tool_call.function.arguments, str):
                                    assert isinstance(  # nosec: B101
                                        accumulated_tool_calls[tool_call.index].function.arguments,
                                        str,
                                    )
                                    accumulated_tool_calls[  # pyright: ignore[reportOperatorIssue]
                                        tool_call.index
                                    ].function.arguments += tool_call.function.arguments

                                else:
                                    assert isinstance(  # nosec: B101
                                        accumulated_tool_calls[tool_call.index].function.arguments,
                                        dict,
                                    )
                                    accumulated_tool_calls[tool_call.index].function.arguments = {
                                        **cast(
                                            dict,
                                            accumulated_tool_calls[
                                                tool_call.index
                                            ].function.arguments,
                                        ),
                                        **tool_call.function.arguments,
                                    }

                    match completion_choice.finish_reason:
                        case None:
                            pass  # continue streaming

                        case "stop":
                            # send tool calls if any
                            for call in accumulated_tool_calls:
                                if not call.function:
                                    continue  # skip partial calls
                                if not call.function.name:
                                    continue  # skip calls with missing names

                                call_identifier: str = call.id or uuid4().hex
                                # send tool requests when ensured that all were completed
                                yield LMMToolRequest(
                                    identifier=call_identifier,
                                    tool=call.function.name,
                                    arguments=json.loads(call.function.arguments)
                                    if isinstance(call.function.arguments, str)
                                    else call.function.arguments,
                                )
                            # send completion chunk if needed
                            yield LMMStreamChunk.of(
                                MultimodalContent.empty,
                                eod=True,
                            )
                            break  # and break the loop

                        case "tool_calls":
                            # send tool calls if any
                            for call in accumulated_tool_calls:
                                if not call.function:
                                    continue  # skip partial calls
                                if not call.function.name:
                                    continue  # skip calls with missing names

                                call_identifier: str = call.id or uuid4().hex
                                ctx.record(
                                    ObservabilityLevel.INFO,
                                    event="lmm.tool_request",
                                    attributes={"lmm.tool": call.function.name},
                                )
                                # send tool requests when ensured that all were completed
                                yield LMMToolRequest(
                                    identifier=call_identifier,
                                    tool=call.function.name,
                                    arguments=json.loads(call.function.arguments)
                                    if isinstance(call.function.arguments, str)
                                    else call.function.arguments,
                                )

                            break  # and break the loop

                        case "length":
                            raise MistralException(
                                "Invalid Mistral completion - exceeded maximum length!",
                                completion_chunk.data,
                            )

                        case "error":
                            raise MistralException(
                                "Mistral completion generation failed!",
                                completion_chunk.data,
                            )

        return ctx.stream(stream)



from typing import Final, Literal

__all__ = (
    "EXCELLENT",
    "FAIR",
    "GOOD",
    "NONE",
    "PERFECT",
    "POOR",
    "EvaluationScoreValue",
    "evaluation_score_value",
)

type EvaluationScoreValue = Literal["none", "poor", "fair", "good", "excellent", "perfect"] | float

NONE: Final[float] = 0.0
POOR: Final[float] = 0.1
FAIR: Final[float] = 0.3
GOOD: Final[float] = 0.5
EXCELLENT: Final[float] = 0.7
PERFECT: Final[float] = 0.9


def evaluation_score_value(  # noqa: PLR0911
    value: EvaluationScoreValue,
    /,
) -> float:
    match value:
        case "none":
            return NONE

        case "poor":
            return POOR

        case "fair":
            return FAIR

        case "good":
            return GOOD

        case "excellent":
            return EXCELLENT

        case "perfect":
            return PERFECT

        case float() as value:
            assert 0 <= value <= 1, "Threshold value has to be in range from 0 to 1"  # nosec: B101
            return value

        case _:
            raise ValueError("Invalid evaluation score value - %s", value)



from typing import Any, Self

from haiway import State, ctx

from draive.evaluation.evaluator import EvaluatorResult, PreparedEvaluator
from draive.evaluation.scenario import PreparedScenarioEvaluator, ScenarioEvaluatorResult
from draive.guardrails.quality.types import GuardrailsQualityException, GuardrailsQualityVerifying
from draive.multimodal import Multimodal, MultimodalContent

__all__ = ("GuardrailsQualityVerification",)


async def _no_verification(
    content: MultimodalContent,
    /,
    **extra: Any,
) -> None:
    pass


class GuardrailsQualityVerification(State):
    @classmethod
    def of(
        cls,
        evaluator: PreparedScenarioEvaluator | PreparedEvaluator,
    ) -> Self:
        async def verifying(
            content: MultimodalContent,
            /,
            **extra: Any,
        ) -> None:
            result: ScenarioEvaluatorResult | EvaluatorResult = await evaluator(content)
            if result.passed:
                return  # verification passed

            match result:
                case ScenarioEvaluatorResult() as result:
                    raise GuardrailsQualityException(
                        reason=result.scenario,
                        content=content,
                        meta=result.meta,
                    )

                case EvaluatorResult() as result:
                    raise GuardrailsQualityException(
                        reason=result.evaluator,
                        content=content,
                        meta=result.meta,
                    )

        return cls(verifying=verifying)

    @classmethod
    async def verify(
        cls,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> None:
        await ctx.state(cls).verifying(MultimodalContent.of(content))

    verifying: GuardrailsQualityVerifying = _no_verification



from haiway import ctx

from draive.instructions import Instruction, InstructionDeclaration
from draive.multimodal import MultimodalContent
from draive.multimodal.tags import MultimodalTagElement
from draive.parameters import DataModel
from draive.stages import Stage

__all__ = (
    "InstructionPreparationAmbiguity",
    "prepare_instruction",
)


class InstructionPreparationAmbiguity(Exception):
    def __init__(
        self,
        *args: object,
        questions: str,
    ) -> None:
        super().__init__(*args)
        self.questions: str = questions


async def prepare_instruction[CaseParameters: DataModel, Result: DataModel | str](
    instruction: InstructionDeclaration | str,
    /,
    *,
    guidelines: str | None = None,
) -> Instruction:
    with ctx.scope("prepare_instruction"):
        ctx.log_info("Preparing instruction...")

        instruction_declaration: InstructionDeclaration
        match instruction:
            case str() as description:
                instruction_declaration = InstructionDeclaration(
                    name="instruction",
                    description=description,
                    arguments=(),
                )

            case declaration:
                instruction_declaration = declaration

        assert instruction_declaration.description is not None  # nosec: B101
        result: MultimodalContent = await Stage.completion(
            f"<USER_TASK>{instruction_declaration.description}</USER_TASK>{_format_variables(instruction_declaration)}",
            instruction=PREPARE_INSTRUCTION.format(
                guidelines=f"\n<GUIDELINES>{guidelines}</GUIDELINES>\n" if guidelines else "",
            ),
        ).execute()

        if parsed := MultimodalTagElement.parse_first("RESULT_INSTRUCTION", content=result):
            ctx.log_info("...instruction preparation finished!")
            return Instruction.of(
                parsed.content.to_str(),
                name=instruction_declaration.name,
                description=instruction_declaration.description,
            )

        elif parsed := MultimodalTagElement.parse_first("QUESTIONS", content=result):
            ctx.log_error("...instruction preparation requires clarification!")
            raise InstructionPreparationAmbiguity(questions=parsed.content.to_str())

        else:
            ctx.log_error("...instruction preparation failed!")
            raise ValueError("Failed to prepare instruction", result)


def _format_variables(
    instruction: InstructionDeclaration,
) -> str:
    if not instruction.arguments:
        return "<TASK_VARIABLES>N/A</TASK_VARIABLES>"

    arguments: str = "\n".join(
        (
            f"- {argument.name}: {argument.description}"
            if argument.description
            else f"- {argument.name}"
        )
        for argument in instruction.arguments
    )

    return f"<TASK_VARIABLES>\n{arguments}\n</TASK_VARIABLES>"


PREPARE_INSTRUCTION: str = """\
You are an expert prompt engineer preparing system instructions for LLMs. Your goal is to create detailed, actionable instructions for completing the described task without any ambiguities.
You should maintain a clear and concise style, ensuring that the instructions are easy to understand and follow.

<INSTRUCTIONS>
You will be given a description of a task. Thoroughly analyze the meaning, context, variables and restrictions to prepare a detailed, actionable instructions required to complete the task.
When the task is ambiguous or lacks important information provide complete set of questions required to clarify.
When the task is not possible to describe say "Task not achievable".
</INSTRUCTIONS>

<STEPS>
Understand the subject
- thoroughly review the task description, understand its purpose, structure, and key details
- identify any nuances, requirements, and the intended outcomes
- break down the task into individual pieces for better understanding
- examine provided context and variables

Describe in detail
- prepare concise and highly precise instructions while retaining essential details
- use parts below and incorporate any relevant elements into the instruction for structural soundness
- name part titles to fit the context of the instruction

<PARTS>
Introduction and role
- provide the context about the role which should be taken on or what goals and overarching tasks you want it to undertake
- it's best to put context early in the body of the instruction
- this section is suitable for most instructions

Tone guidance
- if the tone or manner of delivery significantly affects results quality, include guidance on tone
- this element may not be necessary depending on the task

Detailed task description and rules
- expand on the specific task you want to be done, as well as any rules that must be followed
- this is also where you can give instructions on handling irrelevant, ambiguous or impossible to handle inputs
- this section is mandatory as it contains the actual task description

Examples
- provide an example of an ideal response, encase this in tags, feel free to provide multiple examples
- if you do provide multiple examples, include the context about what it is an example of, and enclose each example in its own set of tags
- make sure to provide examples of common edge cases
- use examples as disambiguation tool, skip it for trivial tasks

Input data to process
- if there is data that needs to be processed within the instruction, include it here within relevant tags
- include multiple pieces of data when needed, but be sure to enclose each in its own set of XML tags
- this section may not be necessary depending on the task

Additional guidance, process description and rules
- for tasks with multiple steps or complex logic, it's good to describe detailed subtasks or encourage to analyze and think before giving an answer
- sometimes, you might have to even state "Before you give your answer..." just to make sure it is done first
- not necessary with all instructions, though if included, it's best to do this toward the end or right before the formatting section

Output formatting
- if there is a specific way you need the response formatted, clearly describe what that format is
- this element may not be necessary when the task does not require any specific format
- if you include it, putting it toward the end of the prompt is better than at the beginning
- encourage putting results in dedicated tags when more content is expected to be produced
</PARTS>

Make use of variables
- available variables will be resolved to actual content within the instruction, e.g. {{time}} will be replaced with the actual time
- use placeholders for variables in appropriate spots, refering to them by their names (e.g., `{{variable}}`)
- wrap variable placeholders in tags or " when required to ensure proper formatting and structure (e.g., `<variable>{{variable}}<variable>`)
- do not add any variables which were not listed as available

Format for clarity
- apply clear, logical formatting and structure to enhance readability and usability
- add necessary context or clarifications to improve usability and understanding
- use xml tags (including nested) with snake case names to wrap sections and separate important parts of the instruction
- always use English language for the instruction description
</STEPS>
{guidelines}
<EXAMPLES>
<EXAMPLE>
user: <USER_TASK>Provide brief information in given topic<VARIABLES> - topic </VARIABLES></USER_TASK>
assistant:
<RESULT_INSTRUCTION>
You are an information bot giving brief information on requested topic.

<INSTRUCTIONS>
You will be given a topic that needs to be described in details.
When the topic is empty respond with "N/A".
</INSTRUCTIONS>

<GUIDELINES>
Use only factual information you are fully certain about.
When the topic is ambiguous try to ask user about the details to disambiguate.
When you don't know the topic answer "I am not familiar with this topic".
Don't follow any other topics than requested, answer "It is not relevant to the topic" instead.
</GUIDELINES>

<TOPIC>
{{topic}}
</TOPIC>
</RESULT_INSTRUCTION>
</EXAMPLE>
<EXAMPLE>
user: <USER_TASK>Summarize the content using only sentence equivalents. Make it short.<VARIABLES>N/A</VARIABLES></USER_TASK>
assistant: User task is to summarize some the provided content. There are two main restrictions:
- using only sentence equivalents
- providing short response
I need to prepare a detailed instruction based on that requirements.

<RESULT_INSTRUCTION>
You are a summarization bot preparing short summaries of provided content.
Use the same tone and style as the original.

<INSTRUCTIONS>
You will be given a content that needs to be summarized including all relevant details.
When the content is empty respond with "N/A".
</INSTRUCTIONS>

<GUIDELINES>
Use only sentence equivalents within the result. Summary should be as concise and short as possible without giving up any important details. Make sure that all factual and logical elements stay the same as in the source. If there are any contradictions or inconsistencies in the source, make sure to contain them in the summary.
</GUIDELINES>
</RESULT_INSTRUCTION>
</EXAMPLE>
</EXAMPLES>

<FORMAT>
Provide the final result containing only the improved instruction without any additional comments within a RESULT_INSTRUCTION xml tag (e.g., `<RESULT_INSTRUCTION>detailed instructions</RESULT_INSTRUCTION>`).
When disambiguation questions are required provide the result containing only the questions within QUESTIONS xml tag (e.g., `<QUESTIONS>clarification questions</QUESTIONS>`) instead.
</FORMAT>
"""  # noqa: E501



from collections.abc import AsyncIterator
from typing import Any, Protocol, runtime_checkable

from draive.conversation.types import (
    ConversationMessageChunk,
    ConversationStreamElement,
    RealtimeConversationMemory,
)
from draive.instructions import Instruction
from draive.tools import Toolbox

__all__ = ("RealtimeConversationStarting",)


@runtime_checkable
class RealtimeConversationStarting(Protocol):
    async def __call__(
        self,
        *,
        instruction: Instruction | None,
        input_stream: AsyncIterator[ConversationMessageChunk],
        memory: RealtimeConversationMemory,
        toolbox: Toolbox,
        **extra: Any,
    ) -> AsyncIterator[ConversationStreamElement]: ...



from typing import cast

from draive.evaluation import EvaluationScore, EvaluationScoreValue, evaluator
from draive.multimodal import Multimodal, MultimodalContent, MultimodalTagElement
from draive.stages import Stage

__all__ = ("coverage_evaluator",)


INSTRUCTION: str = """\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Compare the REFERENCE and the EVALUATED content by carefully examining them, then rate\
 the EVALUATED content using solely a coverage metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is coverage - the extent to which the EVALUATED content includes all\
 the key points from the REFERENCE content.
EVALUATED content with good coverage includes all the important information from\
 the REFERENCE content without omitting critical points.
</EVALUATION_CRITERIA>
{guidelines}
<RATING>
Assign a coverage score using exact name of one of the following values:
- "poor" is very low coverage - the content misses most key points from the reference content.
- "fair" is low coverage - the content includes some key points but omits several important ones.
- "good" is moderate coverage - the content covers most key points but misses a few important details.
- "excellent" is high coverage - the content includes nearly all key points with minor omissions.
- "perfect" is very high coverage - the content comprehensively covers all key points from the reference content.
Use the "none" value for content that cannot be rated at all.
</RATING>

<FORMAT>
The final result containing only the numerical score value HAVE to be put inside a `RESULT`\
 xml tag within the result i.e. `<RESULT>good</RESULT>`.
</FORMAT>
"""  # noqa: E501


@evaluator(name="coverage")
async def coverage_evaluator(
    evaluated: Multimodal,
    /,
    *,
    reference: Multimodal,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore(
            value=0,
            comment="Input was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference was empty!",
        )

    completion: MultimodalContent = await Stage.completion(
        MultimodalContent.of(
            "<REFERENCE>",
            reference,
            "</REFERENCE>\n<EVALUATED>",
            evaluated,
            "</EVALUATED>",
        ),
        instruction=INSTRUCTION.format(
            guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n"
            if guidelines is not None
            else ""
        ),
    ).execute()

    if result := MultimodalTagElement.parse_first(
        "RESULT",
        content=completion,
    ):
        return EvaluationScore.of(
            cast(EvaluationScoreValue, result.content.to_str()),
            comment=completion.to_str(),
        )

    else:
        raise ValueError("Invalid evaluator result:\n%s", completion)



from draive.agents.definition import AgentInvocation, agent
from draive.agents.errors import AgentException
from draive.agents.node import Agent, AgentError, AgentMessage, AgentNode, AgentOutput
from draive.agents.workflow import (
    AgentWorkflow,
    AgentWorkflowIdle,
    AgentWorkflowInput,
    AgentWorkflowInvocation,
    AgentWorkflowOutput,
    workflow,
)

__all__ = (
    "Agent",
    "AgentError",
    "AgentException",
    "AgentInvocation",
    "AgentMessage",
    "AgentNode",
    "AgentOutput",
    "AgentWorkflow",
    "AgentWorkflowIdle",
    "AgentWorkflowInput",
    "AgentWorkflowInvocation",
    "AgentWorkflowOutput",
    "agent",
    "workflow",
)



from draive.embedding.state import ImageEmbedding, TextEmbedding
from draive.embedding.types import Embedded, ValueEmbedding

__all__ = (
    "Embedded",
    "ImageEmbedding",
    "TextEmbedding",
    "ValueEmbedding",
)



from typing import Self

from draive.commons import Meta, MetaValues
from draive.multimodal.media import MediaContent
from draive.multimodal.text import TextContent
from draive.parameters import DataModel

__all__ = ("MetaContent",)


class MetaContent(DataModel):
    @classmethod
    def of(
        cls,
        category: str,
        /,
        *,
        content: TextContent | MediaContent | DataModel | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            category=category,
            content=content,
            meta=Meta.of(meta),
        )

    category: str
    content: TextContent | MediaContent | DataModel | None
    meta: Meta



from asyncio import CancelledError, Task
from collections.abc import AsyncIterator, Sequence
from typing import Any
from uuid import UUID, uuid4

from haiway import AsyncQueue, ctx

from draive.conversation.types import (
    ConversationEvent,
    ConversationMessage,
    ConversationMessageChunk,
    ConversationStreamElement,
    RealtimeConversationMemory,
)
from draive.instructions import Instruction
from draive.lmm import (
    LMMSession,
    LMMSessionEvent,
    LMMSessionOutput,
    LMMStreamChunk,
    LMMStreamInput,
    LMMToolRequest,
    LMMToolResponse,
)
from draive.lmm.types import LMMContextElement
from draive.multimodal import MetaContent
from draive.tools import Toolbox
from draive.utils import Processing, ProcessingEvent

__all__ = ("realtime_conversation",)


async def realtime_conversation(  # noqa: C901, PLR0915
    *,
    instruction: Instruction | None,
    input_stream: AsyncIterator[ConversationMessageChunk],
    memory: RealtimeConversationMemory,
    toolbox: Toolbox,
    **extra: Any,
) -> AsyncIterator[ConversationStreamElement]:
    initial_context: Sequence[LMMContextElement] = tuple(
        message.as_lmm_context_element()
        for message in await memory.recall()  # relying on memory recall correctness
        if isinstance(message, ConversationMessage)  # skip events in LMM context
    )

    input_queue = AsyncQueue[LMMStreamInput]()

    async def process_input() -> None:
        try:
            ctx.log_debug("Starting input processing...")
            # TODO: include guardrails check
            async for element in input_stream:
                if isinstance(element, ConversationEvent):
                    continue  # skip events in LMM context

                input_queue.enqueue(
                    LMMStreamChunk.of(
                        element.content,
                        eod=element.eod,
                    )
                )
                await memory.remember(element)  # add to memory

            input_queue.finish()
            ctx.log_debug("...input processing finished!")

        except CancelledError:
            # just finish input on cancel
            input_queue.finish()

        except BaseException as exc:
            input_queue.finish(exception=exc)

    pending_tool_requests: dict[str, Task[None]] = {}

    async def handle_tool_request(
        tool_request: LMMToolRequest,
        /,
    ) -> None:
        try:
            ctx.log_debug(f"Handling tool request ({tool_request.identifier})...")
            response: LMMToolResponse = await toolbox.respond(tool_request)
            # check if task was not cancelled before passing the result to input
            ctx.check_cancellation()
            # deliver the result directly to input
            input_queue.enqueue(response)

        except CancelledError:
            pass  # just cancelled, ignore

        finally:
            # update pending tools - we have delivered or ignored the result
            del pending_tool_requests[tool_request.identifier]
            ctx.log_debug(f"...tool request ({tool_request.identifier}) handling finished!")

    input_processing_task: Task[None] = ctx.spawn(process_input)

    session_output: AsyncIterator[LMMSessionOutput]
    try:
        ctx.log_debug("Preparing LMM session...")
        session_output = await LMMSession.prepare(
            instruction=instruction,
            input_stream=input_queue,
            initial_context=initial_context,
            tools=toolbox.available_tools(),
        )
        del initial_context

    except BaseException as exc:
        ctx.log_error(
            "...failed to prepare LMM session!",
            exception=exc,
        )
        input_processing_task.cancel()
        raise exc

    output_queue = AsyncQueue[ConversationStreamElement]()
    # current output message identifier to match chunks and events
    output_message_identifier: UUID = uuid4()

    async def report_event(
        event: ProcessingEvent,
        /,
    ) -> None:
        ctx.log_debug("Processing conversation event")
        output_queue.enqueue(
            ConversationEvent.of(
                event.name,
                message_identifier=output_message_identifier,
                identifier=event.identifier,
                content=event.content,
                meta=event.meta,
            )
        )

    with ctx.updated(ctx.state(Processing).updated(event_reporting=report_event)):

        async def process_output() -> None:
            nonlocal output_message_identifier
            try:
                ctx.log_debug("Starting output processing...")
                async for chunk in session_output:
                    match chunk:
                        case LMMStreamChunk() as chunk:
                            output_chunk: ConversationMessageChunk = ConversationMessageChunk.model(
                                chunk.content,
                                message_identifier=output_message_identifier,
                                eod=chunk.eod,
                            )
                            # send to output
                            output_queue.enqueue(output_chunk)
                            if chunk.eod:
                                # start next message when current finished
                                output_message_identifier = uuid4()
                            # and add to memory
                            await memory.remember(output_chunk)

                        case LMMSessionEvent() as event:
                            conversation_event: ConversationEvent
                            match event.category:
                                case "interrupted":
                                    # cancel pending tools on interrupt
                                    for task in pending_tool_requests.values():
                                        task.cancel()

                                    conversation_event = ConversationEvent.of(
                                        "interrupted",
                                        message_identifier=output_message_identifier,
                                        meta=event.meta,
                                    )

                                case "completed":
                                    conversation_event = ConversationEvent.of(
                                        "completed",
                                        message_identifier=output_message_identifier,
                                        meta=event.meta,
                                    )

                                case other:
                                    conversation_event = ConversationEvent.of(
                                        other,
                                        message_identifier=output_message_identifier,
                                        content=MetaContent.of(other),
                                        meta=event.meta,
                                    )

                            # send to output
                            output_queue.enqueue(conversation_event)
                            # and add to memory
                            await memory.remember(conversation_event)

                        case LMMToolRequest() as tool_request:
                            pending_tool_requests[tool_request.identifier] = ctx.spawn(
                                handle_tool_request,
                                tool_request,
                            )

            finally:
                # cancel input reading
                input_processing_task.cancel()
                # cancel pending tools
                for task in pending_tool_requests.values():
                    task.cancel()

                ctx.log_debug("...output processing finished!")

        ctx.spawn(process_output)

    return output_queue



from draive.cohere.client import Cohere
from draive.cohere.config import CohereImageEmbeddingConfig, CohereTextEmbeddingConfig
from draive.cohere.types import CohereException

__all__ = (
    "Cohere",
    "CohereException",
    "CohereImageEmbeddingConfig",
    "CohereTextEmbeddingConfig",
)



from collections.abc import AsyncIterator, Iterable
from itertools import chain
from typing import Any, Literal, cast, overload

from google.genai.types import (
    Candidate,
    Content,
    ContentListUnionDict,
    FinishReason,
    FunctionCallingConfigMode,
    FunctionDeclarationDict,
    GenerateContentResponse,
    Modality,
    SchemaDict,
    SpeechConfigDict,
)
from haiway import ObservabilityLevel, as_list, ctx

from draive.gemini.api import GeminiAPI
from draive.gemini.config import GeminiGenerationConfig
from draive.gemini.lmm import (
    DISABLED_SAFETY_SETTINGS,
    context_element_as_content,
    output_as_response_declaration,
    resoluton_as_media_resulution,
    result_part_as_content_or_call,
    tools_as_tools_config,
)
from draive.gemini.types import GeminiException
from draive.gemini.utils import unwrap_missing
from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContext,
    LMMOutput,
    LMMOutputSelection,
    LMMStreamOutput,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.multimodal.content import MultimodalContent, MultimodalContentElement

__all__ = ("GeminiLMMGeneration",)


class GeminiLMMGeneration(GeminiAPI):
    def lmm(self) -> LMM:
        return LMM(completing=self.lmm_completion)

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        config: GeminiGenerationConfig | None = None,
        output: LMMOutputSelection,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        config: GeminiGenerationConfig | None = None,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    async def lmm_completion(  # noqa: C901, PLR0912, PLR0915
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        config: GeminiGenerationConfig | None = None,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput:
        if stream:
            raise NotImplementedError("gemini streaming is not implemented yet")

        generation_config: GeminiGenerationConfig = config or ctx.state(GeminiGenerationConfig)
        with ctx.scope("gemini_lmm_completion", generation_config):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "lmm.provider": "gemini",
                    "lmm.model": generation_config.model,
                    "lmm.temperature": generation_config.temperature,
                    "lmm.max_tokens": generation_config.max_tokens,
                    "lmm.seed": generation_config.seed,
                    "lmm.tools": [tool["name"] for tool in tools] if tools else None,
                    "lmm.tool_selection": f"{tool_selection}" if tools else None,
                    "lmm.stream": stream,
                    "lmm.output": f"{output}",
                    "lmm.instruction": f"{instruction}",
                    "lmm.context": [element.to_str() for element in context],
                },
            )

            content: ContentListUnionDict = list(
                chain.from_iterable([context_element_as_content(element) for element in context])
            )

            response_schema: SchemaDict | None
            response_mime_type: str | None
            response_modalities: list[Modality] | None
            response_schema, response_modalities, response_mime_type, output_decoder = (
                output_as_response_declaration(output)
            )

            functions: list[FunctionDeclarationDict] | None
            function_calling_mode: FunctionCallingConfigMode | None
            functions, function_calling_mode = tools_as_tools_config(
                tools,
                tool_selection=tool_selection,
            )

            completion: GenerateContentResponse = await self._client.aio.models.generate_content(
                model=generation_config.model,
                config={
                    "temperature": unwrap_missing(generation_config.temperature),
                    "top_p": unwrap_missing(generation_config.top_p),
                    "top_k": unwrap_missing(generation_config.top_k),
                    "max_output_tokens": unwrap_missing(generation_config.max_tokens),
                    "candidate_count": 1,
                    "seed": unwrap_missing(generation_config.seed),
                    "stop_sequences": unwrap_missing(
                        generation_config.stop_sequences,
                        transform=as_list,
                    ),
                    # gemini safety is really bad and often triggers false positive
                    "safety_settings": DISABLED_SAFETY_SETTINGS,
                    "system_instruction": Instruction.formatted(instruction),
                    "tools": [{"function_declarations": functions}] if functions else None,
                    "tool_config": {"function_calling_config": {"mode": function_calling_mode}}
                    if function_calling_mode
                    else None,
                    # prevent google from automatically resolving function calls
                    "automatic_function_calling": {"disable": True, "maximum_remote_calls": None},
                    "media_resolution": resoluton_as_media_resulution(
                        generation_config.media_resolution
                    ),
                    "response_modalities": cast(list[str] | None, response_modalities),
                    "response_mime_type": response_mime_type,
                    "response_schema": response_schema,
                    "speech_config": unwrap_missing(
                        generation_config.speech_voice_name,
                        transform=_speech_config,
                    ),
                },
                contents=content,
            )

            if usage := completion.usage_metadata:
                ctx.record(
                    ObservabilityLevel.INFO,
                    metric="lmm.input_tokens",
                    value=usage.prompt_token_count or 0,
                    unit="tokens",
                    attributes={"lmm.model": generation_config.model},
                )
                ctx.record(
                    ObservabilityLevel.INFO,
                    metric="lmm.input_tokens.cached",
                    value=usage.cached_content_token_count or 0,
                    unit="tokens",
                    attributes={"lmm.model": generation_config.model},
                )
                ctx.record(
                    ObservabilityLevel.INFO,
                    metric="lmm.output_tokens",
                    value=usage.candidates_token_count or 0,
                    unit="tokens",
                    attributes={"lmm.model": generation_config.model},
                )
                ctx.record(
                    ObservabilityLevel.INFO,
                    metric="lmm.output_tokens.thoughts",
                    value=usage.thoughts_token_count or 0,
                    unit="tokens",
                    attributes={"lmm.model": generation_config.model},
                )

            if not completion.candidates:
                raise GeminiException(
                    "Invalid Gemini completion - missing candidates!",
                    completion,
                )

            completion_choice: Candidate = completion.candidates[0]

            match completion_choice.finish_reason:
                case FinishReason.STOP:
                    pass  # process results

                case reason:  # other cases are errors
                    raise GeminiException(
                        f"Gemini completion generation failed! Reason: {reason}",
                        completion,
                    )

            completion_content: Content
            if candidate_content := completion_choice.content:
                completion_content = candidate_content

            else:
                raise GeminiException("Missing Gemini completion content!")

            result_content_elements: list[MultimodalContentElement] = []
            tool_requests: list[LMMToolRequest] = []
            for element in chain.from_iterable(
                result_part_as_content_or_call(part) for part in completion_content.parts or []
            ):
                match element:
                    case LMMToolRequest() as tool_request:
                        tool_requests.append(tool_request)

                    case content_element:
                        result_content_elements.append(content_element)

            lmm_completion: LMMCompletion | None
            if result_content_elements:
                lmm_completion = LMMCompletion.of(
                    output_decoder(MultimodalContent.of(*result_content_elements))
                )

            else:
                lmm_completion = None

            if tool_requests:
                assert tools, "Requesting tool call without tools"  # nosec: B101
                completion_tool_calls = LMMToolRequests(
                    content=lmm_completion.content if lmm_completion else None,
                    requests=tool_requests,
                )

                ctx.record(
                    ObservabilityLevel.INFO,
                    event="lmm.tool_requests",
                    attributes={"lmm.tools": [request.tool for request in tool_requests]},
                )

                return completion_tool_calls

            elif lmm_completion:
                ctx.record(
                    ObservabilityLevel.INFO,
                    event="lmm.completion",
                )
                return lmm_completion

            else:
                raise GeminiException("Invalid Gemini completion, missing content!", completion)


def _speech_config(voice_name: str) -> SpeechConfigDict:
    return {
        "voice_config": {
            "prebuilt_voice_config": {"voice_name": voice_name},
        }
    }



from collections.abc import Collection, Iterable
from types import TracebackType
from typing import Literal, final

from haiway import State

from draive.ollama.api import OllamaAPI
from draive.ollama.embedding import OllamaEmbedding
from draive.ollama.lmm_generation import OllamaLMMGeneration

__all__ = ("Ollama",)


@final
class Ollama(
    OllamaLMMGeneration,
    OllamaEmbedding,
    OllamaAPI,
):
    """
    Access to Ollama services, can be used to prepare various functionalities like lmm.
    """

    __slots__ = ("_features",)

    def __init__(
        self,
        server_url: str | None = None,
        features: Collection[Literal["lmm", "text_embedding"]] | None = None,
    ) -> None:
        super().__init__(server_url=server_url)

        self._features: frozenset[Literal["lmm", "text_embedding"]] = (
            frozenset(features) if features is not None else frozenset(("lmm", "text_embedding"))
        )

    async def __aenter__(self) -> Iterable[State]:
        await self._initialize_client()
        state: list[State] = []
        if "lmm" in self._features:
            state.append(self.lmm())

        if "text_embedding" in self._features:
            state.append(self.text_embedding())

        return state

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._deinitialize_client()



from collections.abc import Callable, Sequence
from re import Pattern
from typing import Literal, cast, overload

from draive.splitters.basic import basic_split_text
from draive.splitters.exhaustive import exhaustive_regex_split_text, exhaustive_split_text

__all__ = ("split_text",)


@overload
def split_text(
    text: str,
    *,
    part_size: int,
    count_size: Callable[[str], int],
    separators: Sequence[str | Pattern[str]] | str | Pattern[str] | None = None,
    part_overlap_size: int | None = None,
    mode: Literal["exhaustive_regex"],
) -> list[str]: ...


@overload
def split_text(
    text: str,
    *,
    part_size: int,
    count_size: Callable[[str], int],
    separators: Sequence[str] | str | None = None,
    part_overlap_size: int | None = None,
    mode: Literal["basic", "exhaustive"] = "basic",
) -> list[str]: ...


def split_text(
    text: str,
    part_size: int,
    count_size: Callable[[str], int],
    separators: Sequence[str | Pattern[str]] | str | Pattern[str] | None = None,
    part_overlap_size: int | None = None,
    mode: Literal[
        "basic",
        "exhaustive",
        "exhaustive_regex",
    ] = "basic",
) -> list[str]:
    match mode:
        case "basic":
            assert not isinstance(separators, Pattern)  # nosec: B101
            assert not any(isinstance(separator, Pattern) for separator in separators or [])  # nosec: B101
            return basic_split_text(
                text=text,
                part_size=part_size,
                count_size=count_size,
                separators=cast(Sequence[str] | str | None, separators),
                part_overlap_size=part_overlap_size,
            )

        case "exhaustive":
            assert not isinstance(separators, Pattern)  # nosec: B101
            assert not any(isinstance(separator, Pattern) for separator in separators or [])  # nosec: B101
            return exhaustive_split_text(
                text=text,
                part_size=part_size,
                count_size=count_size,
                separators=cast(Sequence[str] | str | None, separators),
                part_overlap_size=part_overlap_size,
            )
        case "exhaustive_regex":
            return exhaustive_regex_split_text(
                text=text,
                part_size=part_size,
                count_size=count_size,
                separators=separators,
                part_overlap_size=part_overlap_size,
            )



from draive.lmm.state import LMM, LMMSession
from draive.lmm.types import (
    LMMCompleting,
    LMMCompletion,
    LMMContext,
    LMMContextElement,
    LMMException,
    LMMInput,
    LMMOutput,
    LMMOutputSelection,
    LMMSessionEvent,
    LMMSessionOutput,
    LMMSessionOutputSelection,
    LMMStreamChunk,
    LMMStreamInput,
    LMMStreamOutput,
    LMMToolError,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolResponseHandling,
    LMMToolResponses,
    LMMToolSelection,
    LMMToolSpecification,
)

__all__ = (
    "LMM",
    "LMMCompleting",
    "LMMCompletion",
    "LMMContext",
    "LMMContextElement",
    "LMMException",
    "LMMInput",
    "LMMOutput",
    "LMMOutputSelection",
    "LMMSession",
    "LMMSessionEvent",
    "LMMSessionOutput",
    "LMMSessionOutputSelection",
    "LMMStreamChunk",
    "LMMStreamInput",
    "LMMStreamOutput",
    "LMMToolError",
    "LMMToolRequest",
    "LMMToolRequests",
    "LMMToolResponse",
    "LMMToolResponseHandling",
    "LMMToolResponses",
    "LMMToolSelection",
    "LMMToolSpecification",
)



from collections.abc import Collection, Iterable
from types import TracebackType
from typing import Any, Literal, final

from google.genai.client import HttpOptionsDict  # pyright: ignore[reportPrivateImportUsage]
from haiway import State

from draive.gemini.api import GeminiAPI
from draive.gemini.embedding import GeminiEmbedding
from draive.gemini.lmm_generation import GeminiLMMGeneration
from draive.gemini.lmm_session import GeminiLMMSession
from draive.gemini.tokenization import GeminiTokenization

__all__ = ("Gemini",)


@final
class Gemini(
    GeminiLMMGeneration,
    GeminiLMMSession,
    GeminiEmbedding,
    GeminiTokenization,
    GeminiAPI,
):
    """
    Access to Gemini services, can be used to prepare various functionalities like lmm.
    """

    __slots__ = ("_features",)

    def __init__(
        self,
        api_key: str | None = None,
        http_options: HttpOptionsDict | None = None,
        features: Collection[Literal["lmm", "lmm_session", "text_embedding"]] | None = None,
        **extra: Any,
    ) -> None:
        super().__init__(
            api_key=api_key,
            http_options=http_options,
            **extra,
        )

        self._features: frozenset[Literal["lmm", "lmm_session", "text_embedding"]] = (
            frozenset(features)
            if features is not None
            else frozenset(("lmm", "lmm_session", "text_embedding"))
        )

    async def __aenter__(self) -> Iterable[State]:
        state: list[State] = []
        if "lmm" in self._features:
            state.append(self.lmm())

        if "lmm_session" in self._features:
            state.append(self.lmm_session())

        if "text_embedding" in self._features:
            state.append(self.text_embedding())

        return state

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass



from draive.ollama.client import Ollama
from draive.ollama.config import OllamaChatConfig, OllamaEmbeddingConfig
from draive.ollama.types import OllamaException

__all__ = (
    "Ollama",
    "OllamaChatConfig",
    "OllamaEmbeddingConfig",
    "OllamaException",
)
